# Reader extras
# by Jack Miller
# v1.0

# Designed to put extra interesting content in the reader output.

enabled_extras = ['datetime', 'slashdot', 'authors']

hacks = {}

datetime_attrs = [ 'published_parsed', 'updated_parsed' ]

def datetime_extras(body, extra_content, attrs):
    import time

    if attrs['published_parsed']:
        body = "Published: " + time.asctime(tuple(attrs['published_parsed'])) + "<br />" + body
    elif attrs['updated_parsed']:
        body = "Updated: " + time.asctime(tuple(attrs['updated_parsed'])) + "<br />" + body

    return (body, extra_content)

hacks['datetime'] = ('.*', datetime_attrs, datetime_extras)

# Slashdot (example)

slashdot_attrs = [ 'slash_department' ]

def slashdot_extras(body, extra_content, attrs):
    dept = "From the <strong>%s</strong> department<br />" % attrs["slash_department"]
    return (dept + body, extra_content)

hacks['slashdot'] = ('.*slashdot\\.org.*', slashdot_attrs, slashdot_extras)

# Authors (example)

authors_attrs = [ 'author' ]

def authors_extras(body, extra_content, attrs):
    # Requested attributes will always be in attrs, but they may be empty
    # The daemon returns "" for unknown attributes.

    if attrs['author']:
        return (body + ("<br />By: <strong>%s</strong><br />" % attrs['author']), extra_content)
    return (body, extra_content)

hacks['authors'] = ('.*', authors_attrs, authors_extras)

# If you turn on DEBUG_CONTENT, an entire item and all of its content will be
# appended to the Reader output, to let you look at what is available in an
# item.

DEBUG_CONTENT = False

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_next.hooks import on_hook, remove_hook
from canto_curses.reader import ReaderPlugin
from canto_curses.tagcore import tag_updater
from canto_curses.theme import prep_for_display
import pprint
import re

import logging

log = logging.getLogger("EXTRAS")

class ReaderExtrasPlugin(ReaderPlugin):
    def __init__(self, reader):
        self.plugin_attrs = {'edit_extras' : self.edit_extras}
        if DEBUG_CONTENT:
            self.plugin_attrs['edit_debug'] = self.edit_debug

        self.reader = reader
        self.got_attrs = False
        self.needed_attrs = []
        self.do_funcs = []
        self.setup_hook = False

        # Convert enabled extras' URL regexen into re objects once so we don't
        # convert them all on eval() and we don't have to throw exceptions when
        # we're actually running.

        for extra in enabled_extras:
            hacks[extra] = (re.compile(hacks[extra][0]),) + hacks[extra][1:]

    def on_attributes(self, attributes):
        sel = self.reader.callbacks["get_var"]("reader_item")
        if sel and sel.id in attributes:
            # Sucks that we have to check these, but we can't be sure
            # any particular call is ours.
            for attr in self.needed_attrs:
                if attr not in attributes[sel.id]:
                    break
            else:
                remove_hook("curses_attributes", self.on_attributes)
                self.got_attrs = True
                self.reader.callbacks["set_var"]("needs_refresh", True)
                self.reader.callbacks["release_gui"]()

    def _dofuncs(self, body, extra_content, sel):
        for f in self.do_funcs:
            body, extra_content = f(body, extra_content, sel.content)
        return (body, extra_content)

    def edit_extras(self, body, extra_content):
        sel = self.reader.callbacks["get_var"]("reader_item")
        if not sel:
            return

        # If we have the attrs, or are already waiting, bail.
        if self.got_attrs:
            return self._dofuncs(body, extra_content, sel)
        elif self.setup_hook:
            return (body, extra_content)

        item_url = eval(sel.id)["URL"]

        for extra in enabled_extras:
            url_rgx, needed, func = hacks[extra]
            if url_rgx.match(item_url):
                for a in needed:
                    if a not in sel.content and a not in self.needed_attrs:
                        self.needed_attrs.append(a)
                self.do_funcs.append(func)

        if self.needed_attrs != []:
            on_hook("curses_attributes", self.on_attributes)
            tag_updater.request_attributes(sel.id, self.needed_attrs)
            self.setup_hook = True
            return (body, extra_content)
        else:
            self.got_attrs = True
            return self._dofuncs(body, extra_content, sel)

    # Body is sent through the HTML process, so HTML formatting will be allowed
    # and link tags will be used by goto et. al

    # Extra content is naturally formatted text that may have canto theme escapes

    def edit_debug(self, body, extra_content):
        sel = self.reader.callbacks["get_var"]("reader_item")
        if not sel:
            return

        if not self.got_attrs:
            on_hook("curses_attributes", self.on_attributes)
            tag_updater.request_attributes(sel.id, [])
        else:
            extra_content += '\n\n'
            for k in sel.content.keys():
                line = '[%s]: %s' % (k, sel.content[k])
                extra_content += prep_for_display(line) + '\n'

        return(body, extra_content)



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

CANTO_PROTOCOL_COMPATIBLE = 0.9

from canto_next.client import CantoClient
from canto_next.plugins import try_plugins, set_program
from canto_next.rwlock import alllocks
from canto_next.hooks import call_hook

from .config import config, finalize_eval_settings
from .tagcore import tag_updater, alltagcores
from .gui import CantoCursesGui, GraphicalLog

from threading import Thread
from queue import Queue

import logging

logging.basicConfig(
        format = "%(asctime)s : %(name)s -> %(message)s",
        datefmt = "%H:%M:%S",
        level = logging.INFO
)

log = logging.getLogger("CANTO-CURSES")

import traceback
import locale
import getopt
import signal
import errno
import fcntl
import time
import sys
import os

# It's the CantoCurses class' responsibility to provide the subsequent Gui
# object with a solid foundation with other components. This includes parsing
# command line arguments, starting a canto-daemon instance if necessary, signal
# handling, and wrapping the socket communication.

class CantoCurses(CantoClient):

    def init(self):

        # For good curses behavior.
        locale.setlocale(locale.LC_ALL, '')

        # Used for GUI-signalled death.
        self.pid = os.getpid()
        self.done = False

        # Whether or not to append pid to logfile
        # (debug option)
        self.log_fname_pid = False

        version = "canto-curses " + "0.9.9" + " " + ""
        optl = self.common_args('hl', ["help"], version)
        if optl == -1:
            sys.exit(-1)

        if self.args(optl):
            sys.exit(-1)

        rootlog = logging.getLogger()
        rootlog.setLevel(max(rootlog.level - 10 * self.verbosity,0))

        self.glog_handler = GraphicalLog()

        try:
            if self.port < 0:
                # If we're running locally, ensure daemon is running
                self.start_daemon()
                CantoClient.__init__(self, self.socket_path)
            else:
                CantoClient.__init__(self, None,\
                        port = self.port, address = self.addr)
        except Exception as e:
            log.error("Error: %s" % e)
            sys.exit(-1)

        # __init__ above started one connection, start another
        # for priority stuff.

        self.connect()

        # Make sure we have permissions on the relevant, non-daemon files in
        # the target directory (None of these will be used until we set_log)

        if self.ensure_paths():
            sys.exit(-1)

        self.set_log()
        log.info(version)

        # Evaluate anything in the target /plugins directory.
        set_program("canto-curses")
        self.plugin_errors = try_plugins(self.conf_dir, self.plugin_default, self.disabled_plugins,
                self.enabled_plugins)


    def print_help(self):
        print("USAGE: canto-curses [options]")
        print("\t-h/--help\tThis help")
        print("\t-V/--version\tPrint version")
        print("\t-v/\t\tVerbose logging (for debug)")
        print("\t-D/--dir <dir>\tSet configuration directory.")
        print("\t-l\t\tAppend pid to log file name")
        print("\nPlugin control\n")
        print("\t--noplugins\t\t\t\tDisable plugins")
        print("\t--enableplugins 'plugin1 plugin2...'\tEnable single plugins (overrides --noplugins)")
        print("\t--disableplugins 'plugin1 plugin2...'\tDisable single plugins")
        print("\nNetwork control\n")
        print("NOTE: These should be used in conjunction with SSH port forwarding to be secure\n")
        print("\t-a/--address <IP>\tConnect to this address")
        print("\t-p/--port <port>\tConnect to this port")

    def args(self, optlist):
        for opt, arg in optlist:
            if opt in ["-h", "--help"]:
                self.print_help()
                return 1
            elif opt in ["-l"]:
                self.log_fname_pid = True
        return 0

    def winch(self, a = None, b = None):
        if self.gui.alive:
            self.gui.winch()

    def sigusr1(self, a = None, b = None):
        import threading

        held_locks = {}
        code = {}
        curthreads = threading.enumerate()

        for threadId, stack in sys._current_frames().items():
            name = str(threadId)
            for ct in curthreads:
                if ct.ident == threadId:
                    name = ct.name

            code[name] = ["NAME: %s" % name]
            for filename, lineno, fname, line in traceback.extract_stack(stack):
                code[name].append('FILE: "%s", line %d, in %s' % (filename, lineno, fname))
                if line:
                    code[name].append("  %s" % (line.strip()))

            held_locks[name] = ""
            for lock in alllocks:
                if lock.writer_id == threadId:
                    held_locks[name] += ("%s(w)" % lock.name)
                    continue
                for reader_id, reader_stack in lock.reader_stacks:
                    if reader_id == threadId:
                        held_locks[name] += ("%s(r)" % lock.name)

        for k in code:
            log.info('\n\nLOCKS: %s \n%s' % (held_locks[k], '\n'.join(code[k])))

        log.info("\n\nSTACKS:")
        for lock in alllocks:
            for (reader_id, reader_stack) in lock.reader_stacks:
                log.info("Lock %s (%s readers)" % (lock.name, lock.readers))
                log.info("Lock reader (thread %s):" % (reader_id,))
                log.info(''.join(reader_stack))

            for writer_stack in lock.writer_stacks:
                log.info("Lock %s (%s readers)" % (lock.name, lock.readers))
                log.info("Lock writer (thread %s):" % (lock.writer_id,))
                log.info(''.join(writer_stack))

        log.info("VARS: %s" % config.vars)
        log.info("OPTS: %s" % config.config)

    def child(self, a = None, b = None):
        try:
            while True:
                pid, status = os.waitpid(-1, os.WNOHANG)
                if pid == 0:
                    break
                log.debug("CHLD %d has died: %d", pid, status)
        except Exception as e:
            if e.errno == errno.ECHILD:
                log.debug("CHLD no children?")
            else:
                raise

    def run(self):
        # We want this as early as possible
        signal.signal(signal.SIGUSR1, self.sigusr1)

        # Get config from daemon
        if not config.init(self, CANTO_PROTOCOL_COMPATIBLE):
            print("Invalid daemon version")
            print("Wanted: %s" % CANTO_PROTOCOL_COMPATIBLE)
            print("Got: %s" % config.version)
            sys.exit(-1)
        else:
            log.info("Version check passed: %s" % CANTO_PROTOCOL_COMPATIBLE)

        # Create Tags for each TagCore
        self.gui = CantoCursesGui(self, self.glog_handler)
        tag_updater.init(self)

        # Initial signal setup.
        signal.signal(signal.SIGWINCH, self.winch)
        signal.signal(signal.SIGCHLD, self.child)

        finalize_eval_settings()

        call_hook("curses_start", [])

        if self.plugin_errors:
            log.error("The following error occurred loading plugins:\n\n%s" % self.plugin_errors)

        while self.gui.alive:
            self.gui.tick()
            time.sleep(1)

    def ensure_paths(self):
        if os.path.exists(self.conf_dir):
            if not os.path.isdir(self.conf_dir):
                log.error("Error: %s is not a directory." % self.conf_dir)
                return -1
            if not os.access(self.conf_dir, os.R_OK):
                log.error("Error: %s is not readable." % self.conf_dir)
                return -1
            if not os.access(self.conf_dir, os.W_OK):
                log.error("Error: %s is not writable." % self.conf_dir)
                return -1
        else:
            try:
                os.makedirs(self.conf_dir)
            except Exception as e:
                log.error("Exception making %s : %s" % (self.conf_dir, e))
                return -1
        return self.ensure_files()

    def ensure_files(self):
        logname = "curses-log"
        if self.log_fname_pid:
            logname += ".%d" % os.getpid()

        for f in [ logname ] :
            p = self.conf_dir + "/" + f
            if os.path.exists(p):
                if not os.path.isfile(p):
                    log.error("Error: %s is not a file." % p)
                    return -1
                if not os.access(p, os.R_OK):
                    log.error("Error: %s is not readable." % p)
                    return -1
                if not os.access(p, os.W_OK):
                    log.error("Error: %s is not writable." % p)
                    return -1

        self.log_path = self.conf_dir + "/" + logname

    def set_log(self):
        f = open(self.log_path, "w")
        os.dup2(f.fileno(), sys.stderr.fileno())

    def start(self):
        try:
            self.init()
            self.run()
        except KeyboardInterrupt:
            pass

        except Exception as e:
            tb = traceback.format_exc()
            log.error("Exiting on exception:")
            log.error("\n" + "".join(tb))

        call_hook("curses_exit", [])
        log.info("Exiting.")
        sys.exit(0)

    def __init__(self):
        self.start()



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

# SubThread is just a basic wrapper for a sub connection from the backend that
# dispatches to sub functions based on socket traffic

from threading import Thread, Lock
import traceback
import logging

log = logging.getLogger("SUBTHREAD")

class SubThread(object):
    def init(self, backend):
        self.backend = backend

        # Start up our own connection
        self.conn = backend.connect()
        self.prot_thread = None
        self.alive = False

    def prot_except(self, exception):
        log.error("%s" % exception)

    def prot_errors(self, errors):
        for key in list(errors.keys()):
            for val, error in errors[key]:
                log.error("%s = %s : %s" % (key, val, error))

    def prot_info(self, info):
        log.info("%s" % info)

    def write(self, cmd, args):
        return self.backend.do_write(self.conn, cmd, args)

    def read(self):
        return self.backend.do_read(self.conn)

    def pthread(self):
        self.alive = True

        try:
            while self.alive:
                r = self.read()

                if not r:
                    continue

                # HUP
                if r == 16:
                    self.alive = False
                    break

                cmd, args = r
                protfunc = "prot_" + cmd.lower()
                if hasattr(self, protfunc):
                    getattr(self, protfunc)(args)

                    # For test-suite
                    if hasattr(self.backend, "processed"):
                        self.backend.processed(cmd, args)

                else:
                    log.error("Unknown response?")
                    log.error("%s - %s" % (cmd, args))
        except Exception as e:
            log.error("Thread exception: %s" % (e,))
            log.error(''.join(traceback.format_exc()))

        log.info("Thread exiting - disconnected\nAny further changes will be forgotten!")

    def start_pthread(self):
        self.prot_thread = Thread(target=self.pthread)
        self.prot_thread.daemon = True
        self.prot_thread.start()




# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from html.parser import HTMLParser
import html.entities
import re

from .color import cc

import logging

log = logging.getLogger("HTML")

class CantoHTML(HTMLParser):

    # Reset is used, instead of __init__ so a single
    # instance of the class can parse multiple HTML
    # fragments.

    def reset(self):
        HTMLParser.reset(self)
        self.result = ""
        self.list_stack = []
        self.verbatim = 0

        self.links = []
        self.link_text = ""
        self.link_href = ""
        self.link_open = False

    # unknown_* funnel all tags to handle_tag

    def handle_starttag(self, tag, attrs):
        self.handle_tag(tag, attrs, 1)

    def handle_endtag(self, tag):
        self.handle_tag(tag, {}, 0)

    def handle_data(self, text):
        self.handle_data_clean(text.replace("\\","\\\\",).replace("%","\\%"))

    def handle_data_clean(self, text):
        if self.verbatim <= 0:
            text = text.replace("\n", " ")

        if self.link_open:
            log.debug("adding %s to link_text", text)
            self.link_text += text

        self.result += text

    def convert_charref(self, ref):
        try:
            if ref[0] in ['x','X']:
                c = int(ref[1:], 16)
            else:
                c = int(ref)
        except:
            return ref
        return chr(c)

    def handle_charref(self, ref):
        self.result += self.convert_charref(ref)

    def convert_entityref(self, ref):
        if ref in html.entities.name2codepoint:
            return chr(html.entities.name2codepoint[ref])
        return "[?]"

    def handle_entityref(self, ref):
        self.result += self.convert_entityref(ref)

    # This is the real workhorse of the HTML parser.

    def attr_dict(self, attrs):
        d = {}
        for k, v in attrs:
            d[k] = v
        return d
    
    def handle_tag(self, tag, attrs, open):
        # Convert attrs (list of two-tuples) to dict.
        attrs = self.attr_dict(attrs)

        if tag in ["a"]:
            if open:
                if "href" not in attrs:
                    return
                self.link_open = True
                self.link_href = attrs["href"]
                self.result += cc("reader_link")
            else:
                self.links.append(("link", self.link_href, self.link_text))
                self.link_text = ""
                self.link_href = ""
                self.link_open = False
                self.result += "[" + str(len(self.links)) + "]" + cc.end("reader_link")

        elif tag in ["img"]:
            if open:
                if "src" not in attrs:
                    return
                if "alt" not in attrs:
                    attrs["alt"] = ""
                self.links.append(("image", attrs["src"], attrs["alt"]))
                self.handle_data_clean(cc("reader_image_link") + attrs["alt"] +\
                        "[" + str(len(self.links)) + "]" + cc.end("reader_image_link"))

        elif tag in ["h" + str(x) for x in range(1,7)]:
            if open:
                self.result += "\n%B"
            else:
                self.result += "%b\n"
        elif tag in ["blockquote"]:
            if open:
                self.result += "\n%Q"
            else:
                self.result += "%q\n"
        elif tag in ["pre","code"]:
            if open:
                if tag == "pre":
                    self.result += "\n%Q"
                self.verbatim += 1
            else:
                if tag == "pre":
                    self.result += "%q\n"
                self.verbatim -= 1
        elif tag in ["sup"]:
            if open:
                self.result += "^"
        elif tag in ["p", "br", "div"]:
            self.result += "\n"
        elif tag in ["ul", "ol"]:
            if open:
                self.result += "\n%I"
                self.list_stack.append([tag,0])
            else:
                # Grumble grumble. Bad HTML.
                if len(self.list_stack):
                    self.list_stack.pop()
                self.result += "%i\n"
        elif tag in ["li"]:
            if open:
                self.result += "\n"

                # List item with no start tag, default to ul
                if not len(self.list_stack):
                    self.list_stack.append(["ul",0])

                if self.list_stack[-1][0] == "ul":
                    self.result += "\u25CF "
                else:
                    self.list_stack[-1][1] += 1
                    self.result += str(self.list_stack[-1][1])+ ". "
            else:
                self.result += "\n"

        elif tag in ["i", "small", "em"]:
            if open:
                self.result += cc("reader_italics")
            else:
                self.result += cc.end("reader_italics")
        elif tag in ["b", "strong"]:
            if open:
                self.result += "%B"
            else:
                self.result += "%b"

    def ent_wrapper(self, match):
        return self.convert_entityref(match.groups()[0])

    def char_wrapper(self, match):
        return self.convert_charref(match.groups()[0])

    def convert(self, s):
        # We have this try except because under no circumstances
        # should the HTML parser crash the application. Better
        # handling is done per case in the handler itself so that
        # bad HTML doesn't necessarily lead to garbage output.

        try:
            self.feed(s)
        except Exception as e:
            r = "Error Parsing Content:\n\n"
            r += ("%s" % e)
            l = []
        else:
            r = self.result
            l = self.links

        self.reset()
        return (r,l)

htmlparser = CantoHTML()

html_entity_regex = re.compile("&(\w{1,8});")

def html_entity_convert(s):
    return html_entity_regex.sub(htmlparser.ent_wrapper, s)

char_ref_regex = re.compile("&#([xX]?[0-9a-fA-F]+)[^0-9a-fA-F]")

def char_ref_convert(s):
    return char_ref_regex.sub(htmlparser.char_wrapper, s)



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.hooks import on_hook, remove_hook, unhook_all
from canto_next.plugins import Plugin

from .command import register_commands, register_arg_types, unregister_all, _int_range, _int_check, _string
from .tagcore import tag_updater, alltagcores
from .locks import config_lock
from .guibase import GuiBase
from .reader import Reader
from .tag import Tag, alltags

import logging
import curses
import shlex
import os
import re

log = logging.getLogger("TAGLIST")

# TagList is the class renders a classical Canto list of tags into the given
# panel. It defers to the Tag class for the actual individual tag rendering.
# This is the level at which commands are taken and the backend is communicated
# with.

class TagListPlugin(Plugin):
    pass

class TagList(GuiBase):
    def init(self, pad, callbacks):
        GuiBase.init(self)

        # Drawing information
        self.pad = pad
        self.height, self.width = self.pad.getmaxyx()

        # Callback information
        self.callbacks = callbacks

        # Holster for a list of items for batch operations.
        self.got_items = []

        self.first_sel = None

        self.first_story = None
        self.last_story = None

        self.tags = []

        # Hold config log so we don't miss any new TagCores or get updates
        # before we're ready.

        on_hook("curses_eval_tags_changed", self.on_eval_tags_changed, self)
        on_hook("curses_items_added", self.on_items_added, self)
        on_hook("curses_items_removed", self.on_items_removed, self)
        on_hook("curses_tag_updated", self.on_tag_updated, self)
        on_hook("curses_stories_added", self.on_stories_added, self)
        on_hook("curses_stories_removed", self.on_stories_removed, self)
        on_hook("curses_opt_change", self.on_opt_change, self)
        on_hook("curses_new_tagcore", self.on_new_tagcore, self)
        on_hook("curses_del_tagcore", self.on_del_tagcore, self)

        args = {
            "cursor-offset": ("[cursor-offset]", self.type_cursor_offset),
            "item-list": ("[item-list]: List of item indices (tab complete to show)\n  Simple: 1,3,6,5\n  Ranges: 1-100\n  All: *\n  Selected item: .\n  Domains tag,1,2,3 for 1,2,3 of current tag", self.type_item_list, self.hook_item_list),
            "item-state": ("[item-state]: Any word, can be inverted with minus ex: '-read' or 'marked'", self.type_item_state),
            "tag-list": ("[tag-list]: List of tag indices (tab complete to show)\n  Simple: 1,3,6,5\n  Ranges: 1-100\n  Selected tag: .\n  All: *", self.type_tag_list, self.hook_tag_list),
            # string because tag-item will manually bash in user: prefix
            "user-tag" : ("[user-tag]: Any string, like 'favorite', or 'cool'", self.type_user_tag),
            "category" : ("[category]: Any string, like 'news' or 'comics'", self.type_category),
        }

        base_cmds = {
            "remote delfeed" : (self.cmd_delfeed, ["tag-list"], "Unsubscribe from feeds."),
        }

        nav_cmds = {
            "page-down": (self.cmd_page_down, [], "Move down a page of items"),
            "page-up": (self.cmd_page_up, [], "Move up a page of items"),
            "next-tag" : (self.cmd_next_tag, [], "Scroll to next tag"),
            "prev-tag" : (self.cmd_prev_tag, [], "Scroll to previous tag"),
            "next-marked" : (self.cmd_next_marked, [], "Scroll to next marked item"),
            "prev-marked" : (self.cmd_prev_marked, [], "Scroll to previous marked item"),
            "rel-set-cursor 1": (lambda : self.cmd_rel_set_cursor(1), [], "Next item"),
            "rel-set-cursor -1": (lambda : self.cmd_rel_set_cursor(-1), [], "Previous item"),
        }

        hidden_cmds = {
            "rel-set-cursor": (self.cmd_rel_set_cursor, ["cursor-offset"], "Move the cursor by cursor-offset items"),
        }

        grouping_cmds = {
            "foritems": (self.cmd_foritems, ["item-list"], "Collect items for future commands\n\nAfter a foritems call, subsequent commands that take [item-lists] will use them.\n\nCan be cleared with clearitems."),
            "foritem": (self.cmd_foritem, ["item-list"], "Collect first item for future commands\n\nAfter a foritem call, subsequent commands that take [item-lists] will use the first item given.\n\nCan be cleared with clearitems."),
            "clearitems": (self.cmd_clearitems, [], "Clear collected items (see foritem / foritems)"),
        }

        item_cmds = {
            "goto": (self.cmd_goto, ["item-list"], "Open story links in browser"),
            "reader": (self.cmd_reader, ["item-list"], "Open the built-in reader"),
            "tag-item" : (self.cmd_tag_item, ["user-tag", "item-list"], "Add a tag to individual items"),
            "tags": (self.cmd_tags, ["item-list"], "Show tag of selected items"),
            "item-state": (self.cmd_item_state, ["item-state", "item-list"], "Set item state (i.e. 'item-state read .')"),
            "tag-state": (self.cmd_tag_state, ["item-state", "tag-list"], "Set item state for all items in tag (i.e. 'tag-state read .')"),
        }

        collapse_cmds = {
            "collapse" : (self.cmd_collapse, ["tag-list"], "Collapse tags - reduce the tag's output to a simple status line."),
            "uncollapse" : (self.cmd_uncollapse, ["tag-list"], "Uncollapse tags - show the full content of a tag"),
            "toggle-collapse" : (self.cmd_toggle_collapse, ["tag-list"], "Toggle collapsed state of tags."),
        }


        search_cmds = {
            "search" : (self.cmd_search, ["string"], "Search items for string"),
            "search-regex" : (self.cmd_search_regex, ["string"], "Search items for regex"),
        }

        tag_cmds = {
            "promote" : (self.cmd_promote, ["tag-list"], "Move tags up in the display order (opposite of demote)"),
            "demote" : (self.cmd_demote, ["tag-list"], "Move tags down in the display order (opposite of promote)"),
        }

        tag_group_cmds = {
            "categorize" : (self.cmd_categorize, ["category", "tag-list"], "Categorize a tag"),
            "remove-category" : (self.cmd_remove_category, ["category", "tag-list"], "Remove a tag from a category"),
            "categories" : (self.cmd_categories, ["tag-list"], "Query what categories a tag is in."),
            "show-category" : (self.cmd_show_category, ["category"], "Show only tags in category."),
        }

        register_commands(self, base_cmds, "Base")
        register_commands(self, nav_cmds, "Navigation")
        register_commands(self, hidden_cmds, "hidden")
        register_commands(self, grouping_cmds, "Grouping")
        register_commands(self, item_cmds, "Item")
        register_commands(self, collapse_cmds, "Collapse")
        register_commands(self, search_cmds, "Search")
        register_commands(self, tag_cmds, "Tag")
        register_commands(self, tag_group_cmds, "Tag Grouping")

        register_arg_types(self, args)

        self.plugin_class = TagListPlugin
        self.update_plugin_lookups()

    def die(self):
        log.debug("Cleaning up hooks...")
        unhook_all(self)
        unregister_all(self)

    def tag_by_item(self, item):
        return item.parent_tag

    def tag_by_obj(self, obj):
        if obj.is_tag:
            return obj
        return obj.parent_tag

    # Types return (completion generator, validator)

    # None completions indicates that the help text should be enough (which
    # happens if it's a generic type without bounds)

    def type_cursor_offset(self):
        return (None, _int_check)

    def unhook_item_list(self, vars):
        # Perhaps this should be a separate hook for command completion?
        if "input_prompt" in vars:
            self.callbacks["set_opt"]("story.enumerated", False)
            self.callbacks["release_gui"]()
            remove_hook("curses_var_change", self.unhook_item_list)

    def hook_item_list(self):
        if not self.callbacks["get_opt"]("story.enumerated"):
            self.callbacks["set_opt"]("story.enumerated", True)
            self.callbacks["release_gui"]()
            on_hook("curses_var_change", self.unhook_item_list, self)

    def type_item_list(self):
        all_items = []
        for tag in self.tags:
            if tag.collapsed:
                continue
            for s in tag:
                all_items.append(s)

        domains = { 'all' : all_items }

        syms = { 'all' : {} }
        sel = self.callbacks["get_var"]("selected")

        if sel:
            # If we have a selection, we have a sensible tag domain

            tag = self.tag_by_obj(sel)
            domains['tag']  = [ x for x in tag ]
            syms['tag'] = {}

            if not sel.is_tag:
                syms['tag']['.'] = [ domains['tag'].index(sel) ]
                syms['tag']['*'] = range(0, len(domains['tag']))
                syms['all']['.'] = [ all_items.index(sel) ]
            elif len(sel) > 0:
                syms['tag']['.'] = [ 0 ]
                syms['tag']['*'] = range(0, len(sel))
            else:
                syms['tag']['.'] = []
                syms['tag']['*'] = []
        else:
            syms['all']['.'] = [ ]

        syms['all']['*'] = range(0, len(all_items))

        # if we have items, pass them in, otherwise pass in selected which is the implied context

        fallback = self.got_items[:]
        if fallback == [] and sel and not sel.is_tag:
            fallback = [ sel ]

        return (None, lambda x: _int_range("item", domains, syms, fallback, x))

    def unhook_tag_list(self, vars):
        # Perhaps this should be a separate hook for command completion?
        if "input_prompt" in vars:
            self.callbacks["set_opt"]("taglist.tags_enumerated", False)
            self.callbacks["release_gui"]()
            remove_hook("curses_var_change", self.unhook_tag_list)

    def hook_tag_list(self):
        if not self.callbacks["get_opt"]("taglist.tags_enumerated"):
            self.callbacks["set_opt"]("taglist.tags_enumerated", True)
            self.callbacks["release_gui"]()
            on_hook("curses_var_change", self.unhook_tag_list, self)

    def type_tag_list(self):
        vtags = self.callbacks["get_var"]("taglist_visible_tags")

        domains = { 'all' : vtags }
        syms = { 'all' : {} }

        sel = self.callbacks["get_var"]("selected")

        deftags = []
        if sel and sel.is_tag:
            deftags = [ sel ]
            syms['all']['.'] = [ vtags.index(sel) ]
        elif sel:
            deftags = [ self.tag_by_item(sel) ]
            syms['all']['.'] = [ vtags.index(deftags[0]) ]
        else:
            syms['all']['.'] = [ ]

        syms['all']['*'] = range(0, len(vtags))

        for i, tag in enumerate(vtags):
            if tag.tag.startswith("maintag:"):
                syms['all'][tag.tag[8:]] = [ i ]

        return (None, lambda x: _int_range("tag", domains, syms, deftags, x))

    # This will accept any state, but should offer some completions for sensible ones

    def type_item_state(self):
        return (["read","marked","-read","-marked"], lambda x : (True, x))

    def on_new_tagcore(self, tagcore):
        log.debug("Instantiating Tag() for %s", tagcore.tag)
        Tag(tagcore, self.callbacks)
        self.callbacks["set_var"]("needs_refresh", True)

    def on_del_tagcore(self, tagcore):
        log.debug("taglist on_del_tag")
        for tagobj in alltags:
            if tagobj.tag == tagcore.tag:
                tagobj.die()

        self.callbacks["set_var"]("needs_refresh", True)

    # We really shouldn't care about item being added (it's a TagCore event)
    # but we do need to release the gui thread so that it can handle sync
    # caused by an empty Tag's TagCore getting items.

    def on_items_added(self, tagcore, items):
        self.callbacks["release_gui"]()

    def on_items_removed(self, tagcore, items):
        self.callbacks["release_gui"]()

    def on_tag_updated(self, tagcore):
        self.callbacks["release_gui"]()

    def on_eval_tags_changed(self):
        self.callbacks["force_sync"]()
        self.callbacks["release_gui"]()

    # Called with sync_lock, so we are unrestricted.

    def on_stories_added(self, tag, items):
        # Items being added implies we need to remap them
        self.callbacks["set_var"]("needs_refresh", True)

    # Called with sync_lock, so we are unrestricted.

    def on_stories_removed(self, tag, items):
        # Items being removed implies we need to remap them.
        self.callbacks["set_var"]("needs_refresh", True)

    def on_opt_change(self, conf):
        if "taglist" not in conf or "search_attributes" not in conf["taglist"]:
            return

        log.info("Fetching any needed search attributes")

        need_attrs = {}
        sa = self.callbacks["get_opt"]("taglist.search_attributes")

        # Make sure that we have all attributes needed for a search.
        for tag in alltagcores:
            for item in tag:
                tag_updater.need_attributes(item, sa)

    def cmd_goto(self, items):
        log.debug("GOTO: %s", items)
        self._goto([item.content["link"] for item in items])

    def cmd_tag_state(self, state, tags):
        attributes = {}
        for tag in tags:
            for item in tag:
                if item.handle_state(state):
                    attributes[item.id] = { "canto-state" : item.content["canto-state"] }

        if attributes:
            tag_updater.set_attributes(attributes)

    # item-state: Add/remove state for multiple items.

    def cmd_item_state(self, state, items):
        attributes = {}
        for item in items:
            if item.handle_state(state):
                attributes[item.id] = { "canto-state" : item.content["canto-state"] }

        if attributes:
            tag_updater.set_attributes(attributes)

    # tag-item : Same as above, with tags.

    def cmd_tag_item(self, tag, items):
        # Proper prefix
        if tag[0] in '-%':
            tag = tag[0] + "user:" + tag[1:]
        else:
            tag = "user:" + tag

        attributes = {}
        for item in items:
            if item.handle_tag(tag):
                attributes[item.id] = { "canto-tags" : item.content["canto-tags"] }

        if attributes:
            tag_updater.set_attributes(attributes)

    def cmd_tags(self, items):
        for item in items:
            if "title" in item.content:
                log.info("'%s' in tags:\n" % item.content["title"])

            log.info(item.parent_tag.tag)

            if "canto-tags" in item.content:
                for tag in item.content["canto-tags"]:
                    if tag.startswith("user:"):
                        log.info(tag[5:])
                    else:
                        log.info(tag)

    def _iterate_forward(self, start):
        ns = start.next_sel
        o = start

        lines = 0

        # No next item, bail.

        if not ns:
            return (None, lines)

        # Force changes to all objects between
        # start and next sel.

        while o and o != ns:
            lines += o.lines(self.width)
            o = o.next_obj

        return (ns, lines)

    def _iterate_backward(self, start):
        ps = start.prev_sel
        o = start

        lines = 0

        # No prev item, bail.

        if not ps:
            return (None, lines)

        # Force changes to all objects between
        # start and prev sel.

        while o and o != ps:
            o = o.prev_obj
            lines += o.lines(self.width)

        return (ps, lines)

    def cmd_rel_set_cursor(self, relidx):
        sel = self.callbacks["get_var"]("selected")
        if sel:
            target_idx = sel.sel_offset + relidx
            curpos = sel.curpos

            if target_idx < 0:
                target_idx = 0

            while sel.sel_offset != target_idx:
                if target_idx < sel.sel_offset and sel.prev_sel:
                    sel, lines = self._iterate_backward(sel)
                    curpos -= lines
                elif target_idx > sel.sel_offset and sel.next_sel:
                    sel, lines = self._iterate_forward(sel)
                    curpos += lines
                else:
                    break
            self._set_cursor(sel, curpos)
        else:
            self._set_cursor(self.first_sel, 0)

    def _set_cursor(self, item, window_location):
        # May end up as None
        sel = self.callbacks["get_var"]("selected")

        if sel:
            sel.unselect()

        self.callbacks["set_var"]("selected", item)

        if item:

            conf = self.callbacks["get_conf"]()
            curstyle = conf["taglist"]["cursor"]

            # Convert window position for absolute positioning, edge
            # positioning uses given window_location.

            if curstyle["type"] == "bottom":
                window_location = 0
            elif curstyle["type"] == "middle":
                window_location = int((self.height - 1) / 2)
            elif curstyle["type"] == "top":
                window_location = self.height - 1

            # If the tag header is larger than the edge, the scroll will never
            # be triggered (redraw resets screen position to keep items visible
            # despite the tag header).

            tag = self.tag_by_obj(item)

            wl_top = max(curstyle["edge"], tag.lines(self.width))

            # Similarly, if the current item is larger than the (edge + 1), the
            # scroll won't be triggered, so we take the max edge there too.

            wl_bottom = (self.height - 1) - max(curstyle["edge"], item.lines(self.width))

            if window_location > wl_bottom:
                if curstyle["scroll"] == "scroll":
                    window_location = wl_bottom
                elif curstyle["scroll"] == "page":
                    window_location = wl_top
            elif window_location < wl_top:
                if curstyle["scroll"] == "scroll":
                    window_location = wl_top
                elif curstyle["scroll"] == "page":
                    window_location = wl_bottom

            self.callbacks["set_var"]("target_obj", item)
            self.callbacks["set_var"]("target_offset", window_location)
            item.select()
        else:
            self.callbacks["set_var"]("target_obj", self.first_sel)
            if self.first_sel:
                self.callbacks["set_var"]("target_offset", self.first_sel.curpos)

    # foritems gets a valid list of items by index.

    def cmd_foritems(self, items):
        self.got_items = items

    def cmd_foritem(self, items):
        if len(items) > 0:
            self.got_items = [ items[0] ]
        else:
            self.got_items = []

    # clearitems clears all the items set by foritems.

    def cmd_clearitems(self):
        log.debug("Clearing ITEMS!")
        self.got_items = []

    def cmd_page_up(self):
        target_offset = self.callbacks["get_var"]("target_offset")
        target_obj = self.callbacks["get_var"]("target_obj")
        sel = self.callbacks["get_var"]("selected")

        # No items, forget about it
        if not target_obj:
            return

        scroll = self.height - 1

        if sel:
            while scroll > 0 and sel.prev_sel:
                pstory = sel.prev_sel
                while sel != pstory:
                    scroll -= sel.lines(self.width)
                    sel = sel.prev_obj

            self._set_cursor(sel, target_offset)
        else:
            while scroll > 0 and target_obj.prev_obj:
                target_obj = target_obj.prev_obj
                scroll -= target_obj.lines(self.width)

            self.callbacks["set_var"]("target_obj", target_obj)
            self.callbacks["set_var"]("target_offset", target_offset)
            self.callbacks["set_var"]("needs_redraw", True)

    def cmd_page_down(self):
        target_offset = self.callbacks["get_var"]("target_offset")
        target_obj = self.callbacks["get_var"]("target_obj")
        sel = self.callbacks["get_var"]("selected")

        # No items, forget about it.
        if not target_obj:
            return

        scroll = self.height - 1

        if sel:
            while scroll > 0 and sel.next_sel:
                if scroll < sel.lines(self.width):
                    break

                nstory = sel.next_sel
                while sel != nstory:
                    scroll -= sel.lines(self.width)
                    sel = sel.next_obj

            self._set_cursor(sel, target_offset)
        else:
            while scroll > 0 and target_obj.next_obj:
                scroll -= target_obj.lines(self.width)
                if scroll < 0:
                    break
                target_obj = target_obj.next_obj

            self.callbacks["set_var"]("target_obj", target_obj)
            self.callbacks["set_var"]("target_offset", 0)
            self.callbacks["set_var"]("needs_redraw", True)

    def cmd_next_tag(self):
        sel = self.callbacks["get_var"]("selected")

        if not sel:
            return self._set_cursor(self.first_sel, 0)

        target_offset = self.callbacks["get_var"]("target_offset")

        tag = self.tag_by_obj(sel)

        while sel and self.tag_by_obj(sel) == tag:
            if sel.next_sel == None:
                break
            sel = sel.next_sel

        self._set_cursor(sel, target_offset)

    def cmd_prev_tag(self):
        sel = self.callbacks["get_var"]("selected")

        if not sel:
            return self._set_cursor(self.first_sel, 0)

        target_offset = self.callbacks["get_var"]("target_offset")

        tag = self.tag_by_obj(sel)

        while sel and self.tag_by_obj(sel) == tag:
            if sel.prev_sel == None:
                break
            sel = sel.prev_sel

        if sel:
            newtag = self.tag_by_obj(sel)
            if newtag.collapsed:
                sel = newtag
            else:
                sel = newtag[0]

        self._set_cursor(sel, target_offset)

    def cmd_reader(self, items):
        self.callbacks["set_var"]("reader_item", items[0])
        self.callbacks["set_var"]("reader_offset", 0)
        self.callbacks["add_window"](Reader)

    def cmd_promote(self, tags):
        for tag in tags:

            log.debug("Promoting %s\n", tag.tag)

            # Refetch because a promote call will cause our eval_tag hook to
            # recreate visible_tags.

            visible_tags = self.callbacks["get_var"]("taglist_visible_tags")

            curidx = visible_tags.index(tag)

            # Obviously makes no sense on top tag.
            if curidx == 0:
                return

            # Re-order tags and update internal list order.
            self.callbacks["switch_tags"](tag.tag, visible_tags[curidx - 1].tag)

        self.callbacks["set_var"]("needs_refresh", True)

    def cmd_demote(self, tags):
        for tag in tags:

            log.debug("Demoting %s\n", tag.tag)

            visible_tags = self.callbacks["get_var"]("taglist_visible_tags")

            # Obviously makes no sense on bottom or only tag.
            if tag == visible_tags[-1] or len(visible_tags) == 1:
                return

            curidx = visible_tags.index(tag)
            self.callbacks["switch_tags"](tag.tag, visible_tags[curidx + 1].tag)

        self.callbacks["set_var"]("needs_refresh", True)

    def _collapse_tag(self, tag):
        log.debug("Collapsing %s\n", tag.tag)

        # If we're collapsing the selection, select
        # the tag instead.
        s = self.callbacks["get_var"]("selected")
        if s and s in tag:
            toffset = self.callbacks["get_var"]("target_offset")
            self._set_cursor(tag, toffset) 

        self.callbacks["set_tag_opt"](tag.tag, "collapsed", True)

    def cmd_collapse(self, tags):
        for tag in tags:
            self._collapse_tag(tag)

    def _uncollapse_tag(self, tag):
        log.debug("Uncollapsing %s\n", tag.tag)

        # If we're uncollapsing the selected tag,
        # go ahead and select the first item.

        s = self.callbacks["get_var"]("selected")
        if s and tag == s and len(tag) != 0:
            toffset = self.callbacks["get_var"]("target_offset") + tag.lines(self.width)
            self._set_cursor(tag[0], toffset)

        self.callbacks["set_tag_opt"](tag.tag, "collapsed", False)

    def cmd_uncollapse(self, tags):
        for tag in tags:
            self._uncollapse_tag(tag)

    def cmd_toggle_collapse(self, tags):
        for tag in tags:
            if self.callbacks["get_tag_opt"](tag.tag, "collapsed"):
                self._uncollapse_tag(tag)
            else:
                self._collapse_tag(tag)

    def search(self, regex):
        try:
            rgx = re.compile(regex)
        except Exception as e:
            self.callbacks["set_var"]("error_msg", e)
            return

        story = self.first_story
        terms = self.callbacks["get_opt"]("taglist.search_attributes")

        while story:
            for t in terms:

                # Shouldn't happen unless a search happens before
                # the daemon can respond to the ATTRIBUTES request.

                if t not in story.content:
                    continue

                if rgx.match(story.content[t]):
                    story.mark()
                    break
            else:
                story.unmark()

            story = story.next_story

        self.callbacks["set_var"]("needs_redraw", True)

    def cmd_search(self, term):
        if not term:
            term = self.callbacks["input"]("search:", False)
        if not term:
            return

        rgx = ".*" + re.escape(term) + ".*"
        return self.search(rgx)

    def cmd_search_regex(self, term):
        if not term:
            term = self.callbacks["input"]("search-regex:", False)
        if not term:
            return
        return self.search(term)

    def cmd_next_marked(self):
        start = self.callbacks["get_var"]("selected")

        # This works for tags and stories alike.
        if start:
            cur = start.next_story
        else:
            start = self.first_story
            cur = start

        # There's nothing to search
        if not cur:
            return

        curpos = cur.curpos

        while not cur or not cur.marked:
            # Wrap to top
            if cur == None:
                cur = self.first_story
                curpos = self.first_story.curpos
            else:
                cur, lines = self._iterate_forward(cur)
                curpos += lines

            # Make sure we don't infinite loop.
            if cur == start:
                if not cur.marked:
                    self.callbacks["set_var"]\
                            ("info_msg", "No marked items.")
                break

        self._set_cursor(cur, curpos)

    def cmd_prev_marked(self):
        start = self.callbacks["get_var"]("selected")

        # This works for tags and stories alike.
        if start:
            cur = start.prev_story
        else:
            start = self.last_story
            cur = start

        # There's nothing to search
        if not cur:
            return

        curpos = cur.curpos

        while not cur or not cur.marked:
            # Wrap to bottom
            if cur == None:
                cur = self.last_story
                curpos = self.last_story.curpos
            else:
                cur, lines = self._iterate_backward(cur)
                curpos -= lines

            # Make sure we don't infinite loop.
            if cur == start:
                self.callbacks["set_var"]("info_msg", "No marked items.")
                break

        self._set_cursor(cur, curpos)

    def type_user_tag(self):
        utags = []
        for tag in alltagcores:
            if tag.tag.startswith("user:"):
                utags.append(tag.tag[5:])

        return (utags, lambda x : (True, x))

    def type_category(self):
        def category_validator(x):
            if x.lower() == "none":
                return (True, None)
            else:
                return (True, x)

        categories = []
        for tag in alltagcores:
            if tag.tag.startswith("category:"):
                categories.append(tag.tag[9:])

        return (categories, category_validator)

    def cmd_categorize(self, category, tags):
        if not category:
            return
        for tag in tags:
            tc = self.callbacks["get_tag_conf"](tag.tag)

            fullcat = "category:" + category
            if fullcat not in tc["extra_tags"]:
                tc["extra_tags"].append(fullcat)
                self.callbacks["set_tag_conf"](tag.tag, tc)
                log.info("%s is now in category %s" % (tag, category))

    def cmd_remove_category(self, category, tags):
        if not category:
            return
        for tag in tags:
            tc = self.callbacks["get_tag_conf"](tag.tag)

            fullcat = "category:" + category
            if fullcat in tc["extra_tags"]:
                tc["extra_tags"].remove(fullcat)
                self.callbacks["set_tag_conf"](tag.tag, tc)
                log.info("%s is no longer in category %s" % (tag, category))

    def cmd_categories(self, tags):
        for tag in tags:
            tc = self.callbacks["get_tag_conf"](tag.tag)
            categories = [ x[9:] for x in tc["extra_tags"] if x.startswith("category:")]
            if categories == []:
                log.info("%s - No categories" % tag)
            else:
                log.info("%s - %s" % (tag, " ".join(categories)))

        popped_cats = []
        for tag in alltagcores:
            if tag.tag.startswith("category:"):
                popped_cats.append(tag.tag[9:])

        if popped_cats:
            log.info("\nAvailable categories:")
            for cat in popped_cats:
                log.info(cat)
        else:
            log.info("\nNo categories available.")

    def cmd_show_category(self, category):
        if category:
            tag_updater.transform("categories", "InTags(\'" + shlex.quote("category:" + category) + "\')")
        else:
            tag_updater.transform("categories", "None")
        tag_updater.update()

    def cmd_delfeed(self, tags):
        for tag in tags:
            if tag.tag.startswith("maintag:"):
                self._remote_argv(["canto-remote", "delfeed", tag.tag[8:]])
            else:
                log.info("tag %s is not a feed tag")

    def update_tag_lists(self):
        curtags = self.callbacks["get_var"]("curtags")
        self.tags = []

        # Make sure to honor the order of tags in curtags.

        for tag in curtags:
            for tagobj in alltags:
                if tagobj.tag == tag:
                    self.tags.append(tagobj)

        # If selected is stale (i.e. its tag was deleted, the item should stick
        # around in all other cases) then unset it.

        sel = self.callbacks["get_var"]("selected")
        tobj = self.callbacks["get_var"]("target_obj")

        if sel and ((sel.is_tag and sel not in self.tags) or (not sel.is_tag and sel.is_dead)):
            log.debug("Stale selection")
            self.callbacks["set_var"]("selected", None)

        if tobj and ((tobj.is_tag and tobj not in self.tags) or (not tobj.is_tag and tobj.is_dead)):
            log.debug("Stale target obj")
            self.callbacks["set_var"]("target_obj", None)
            self.callbacks["set_var"]("target_offset", 0)

        hide_empty = self.callbacks["get_opt"]("taglist.hide_empty_tags")

        cur_item_offset = 0
        cur_sel_offset = 0
        t = []

        for i, tag in enumerate(self.tags):
            if hide_empty and len(tag) == 0:
                continue

            # Update index info
            tag.set_item_offset(cur_item_offset)
            tag.set_sel_offset(cur_sel_offset)
            tag.set_tag_offset(i)
            tag.set_visible_tag_offset(len(t))

            if self.callbacks["get_tag_opt"](tag.tag, "collapsed"):
                cur_sel_offset += 1
            else:
                cur_sel_offset += len(tag)
                cur_item_offset += len(tag)

            t.append(tag)

        self.callbacks["set_var"]("taglist_visible_tags", t)

    def update_target_obj(self):
        # Set initial target_obj if none already set, or if it's stale.

        target_obj = self.callbacks["get_var"]("target_obj")

        if target_obj:
            return

        vistags = self.callbacks["get_var"]("taglist_visible_tags")

        if vistags:
            self.callbacks["set_var"]("target_obj", vistags[0])
            self.callbacks["set_var"]("target_offset", 0)
        else:
            self.callbacks["set_var"]("target_obj", None)
            self.callbacks["set_var"]("target_offset", 0)

    # Refresh updates information used to render the objects.
    # Effectively, we build a doubly linked list out of all
    # of the objects by setting obj.prev_obj and obj.next_obj.

    def refresh(self):

        log.debug("Taglist REFRESH!\n")

        self.update_tag_lists()
        self.update_target_obj()

        self.first_story = None

        prev_obj = None
        prev_story = None
        prev_sel = None

        for tag in self.callbacks["get_var"]("taglist_visible_tags"):
            tag.curpos = self.height

            tag.prev_obj = prev_obj
            tag.next_obj = None

            tag.prev_story = prev_story
            tag.next_story = None

            tag.prev_sel = prev_sel
            tag.next_sel = None

            if prev_obj != None:
                prev_obj.next_obj = tag

            prev_obj = tag

            # Collapsed tags (with items) skip stories.
            if self.callbacks["get_tag_opt"](tag.tag, "collapsed"):
                if prev_sel:
                    prev_sel.next_sel = tag
                prev_sel = tag
                continue

            for story in tag:
                story.curpos = self.height

                if not self.first_story:
                    self.first_story = story

                story.prev_obj = prev_obj
                story.next_obj = None
                prev_obj.next_obj = story
                prev_obj = story

                if prev_story != None:
                    prev_story.next_story = story
                story.prev_story = prev_story
                story.next_story = None
                prev_story = story

                # We want next_story to be accessible from all objects, so head
                # back and set it for any without one, even if it wasn't the
                # last story object (i.e. if it's a tag)

                cur = story.prev_obj
                while cur and cur.next_story == None:
                    cur.next_story = story
                    cur = cur.prev_obj

                if prev_sel != None:
                    prev_sel.next_sel = story
                story.prev_sel = prev_sel
                story.next_sel = None
                prev_sel = story

                # Keep track of last story.
                self.last_story = story

        self.callbacks["set_var"]("needs_redraw", True)

    # curpos - position in visible windown, can be negative
    # main_offset - starting line from top of pad

    def _partial_render(self, obj, main_offset, curpos, footer = False):
        lines = obj.pads(self.width)
        pad = obj.pad

        if footer:
            lines = obj.footlines
            pad = obj.footpad

        draw_lines = lines

        if curpos + lines > 0:
            start = 0

            # If we're crossing the boundary to onscreen
            # trim render window.
            if curpos < 0:
                start = -1 * curpos
                draw_lines += curpos

            # If we're crossing the boundary to offscreen
            # trim render window.
            if main_offset + draw_lines > self.height:
                draw_lines = self.height - main_offset

            if draw_lines:
                pad.overwrite(self.pad, start, 0, main_offset, 0,
                        main_offset + (draw_lines - 1), self.width - 1)
                return (main_offset + draw_lines, curpos + lines)

        return (main_offset, curpos + lines)

    def redraw(self):
        log.debug("Taglist REDRAW (%s)!\n", self.width)
        self.pad.erase()

        target_obj = self.callbacks["get_var"]("target_obj")
        target_offset = self.callbacks["get_var"]("target_offset")

        # Bail if we have no item.

        if target_obj == None:
            self.pad.addstr("All tags empty.")
            self.callbacks["refresh"]()
            return

        # Step 0. Bounding. Make sure we're trying to render the
        # item to a place it's visible.

        # If we're trying to render the target_obj to a screen
        # position less then the length of it's tag header, then
        # we'd overwrite on writing the floating header, so adjust
        # the target_offset.

        if not target_obj.is_tag:
            tag = target_obj.parent_tag
            tl = tag.lines(self.width)
            if target_offset < tl:
                target_offset = tl
        elif target_offset < 0:
            target_offset = 0

        # If we're trying to render too close to the bottom, we also
        # need an adjustment.

        tol = target_obj.lines(self.width)
        if target_offset > ((self.height - 1) - tol):
            target_offset = (self.height - 1) - tol

        # Step 1. Find first object based on target_obj and target_offset,
        # This will cause any changes to be resolved for objects on screen
        # before and including the target object.

        obj = target_obj
        curpos = target_offset
        top_adjusted = False

        while curpos > 0:
            if obj.prev_obj:
                curpos -= obj.prev_obj.lines(self.width)
                obj = obj.prev_obj

            # If there aren't enough items to render before this item and
            # get to the top, adjust offset
            else:
                top_adjusted = True
                target_offset -= curpos
                curpos = 0

        # Step 2. Adjust offset, if necessary, to keep blank space from
        # the bottom of the list. This also causes any changes to be resolved
        # for objects on screen after the target object.

        last_obj = target_obj
        last_off = target_offset

        while last_off < (self.height - 1):
            if last_obj:
                last_off += last_obj.lines(self.width)
                last_obj = last_obj.next_obj

            # Not enough items to render after our item,
            # adjust offset. Unfortunately, this means that
            # we need to refigure out everything above, so
            # we recurse, but as long as we haven't top_adjusted
            # we should only ever have a single level of
            # recursion and none of the refresh work we've done
            # at this level has been wasted.

            elif not top_adjusted:
                rem = (self.height - 1) - last_off
                self.callbacks["set_var"]("target_offset", target_offset + rem)
                self.redraw()
                return
            else:
                break

        # Any adjustments should be reflected.
        self.callbacks["set_var"]("target_offset", target_offset)

        # Step 3. Update self.first_sel. This is useful for making
        # initial selection based on the current screen position.
        # If there are only tags on screen, first_sel could be None

        self.first_sel = obj
        while self.first_sel.is_tag:

            if self.callbacks["get_tag_opt"](obj.tag, "collapsed"):
                break

            # We use obj instead of sel here because next_sel will only be set
            # if the current object is selectable, which it isn't if it's not
            # collapsed.

            if self.first_sel.next_obj:
                self.first_sel = self.first_sel.next_obj
            else:
                break

        # Step 4. Render.

        rendered_header = False
        w_offset = 0

        while obj != None:
            # Refresh if necessary, update curpos for scrolling.
            obj.lines(self.width)
            obj.curpos = curpos

            # Copy item into window
            w_offset, curpos = self._partial_render(obj, w_offset, curpos)

            # If we're at the end of a list, or the next item is a tag we need
            # to render the tag footer for the current tag.

            # Render floating header, if we've covered enough ground.

            if not rendered_header and curpos > 0:
                tag = self.tag_by_obj(obj)

                if curpos >= tag.lines(self.width):
                    self._partial_render(tag, 0, 0)
                    rendered_header = True

            # Do this before the floating header so that if no items are going
            # to be visible, the header is still displayed without the close.

            obj.extra_lines = 0

            if not obj.next_obj or obj.next_obj.is_tag:
                if obj.is_tag:
                    tag = obj
                else:
                    tag = self.tag_by_item(obj)
                    tag.lines(self.width)
                    obj.extra_lines = tag.footlines

                w_offset, curpos = self._partial_render(tag, w_offset, curpos, True)

                # Set this because if we don't have room above the footer for
                # the header (implied by this block executing with
                # rendered_header == False), then actually rendering one looks
                # broken.

                rendered_header = True

            if w_offset >= self.height:
                break

            obj = obj.next_obj

        self.callbacks["refresh"]()

    def is_input(self):
        return False

    def get_opt_name(self):
        return "taglist"

    def get_height(self, mheight):
        return mheight

    def get_width(self, mwidth):
        return mwidth



# Canto Default Theme

# Defined as a plugin to use as a base for other themes.

FORCE_COLORS = False
FORCE_STYLE = False

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_curses.story import StoryPlugin
from canto_curses.tag import TagPlugin
from canto_curses.theme import prep_for_display
from canto_curses.color import cc

cmds = []

if FORCE_COLORS:
    cmds.append("reset-config color")
if FORCE_STYLE:
    cmds.append("reset-config style")

class CantoThemeStoryDefault(StoryPlugin):
    def __init__(self, story):
        self.story = story
        self.plugin_attrs = { "eval" : self.eval }

    def eval(self):
        story = self.story
        s = ""

        if "read" in story.content["canto-state"]:
            s += cc("read")
        else:
            s += cc("unread")

        if story.marked:
            s += cc("marked") + "[*]"

        if story.selected:
            s += cc("selected")

        s += prep_for_display(story.content["title"])

        if story.selected:
            s += cc.end("selected")

        if story.marked:
            s += cc.end("marked")

        if "read" in story.content["canto-state"]:
            s += cc.end("read")
        else:
            s += cc.end("unread")

        return s

class CantoThemeTagDefault(TagPlugin):
    def __init__(self, tag):
        self.tag = tag
        self.plugin_attrs = { "eval" : self.eval }

    def eval(self):
        tag = self.tag

        # Make sure to strip out the category from category:name
        str_tag = tag.tag.split(':', 1)[1]

        unread = len([s for s in tag\
                if "canto-state" not in s.content or\
                "read" not in s.content["canto-state"]])

        s = ""

        if tag.selected:
            s += cc("selected")

        if tag.collapsed:
            s += "[+]"
        else:
            s += "[-]"

        s += " " + str_tag + " "

        s += "[" + cc("unread") + str(unread) + cc.end("unread") + "]"

        if tag.updates_pending:
            s += " [" + cc("pending") + str(tag.updates_pending) + cc.end("pending") + "]"

        if tag.selected:
            s += cc.end("selected")

        return s

# Stolen from autocmd.py, but simple enough to copy instead of introducing a
# dependency.

from canto_curses.gui import GuiPlugin
from canto_next.hooks import on_hook

class AutoCmdGui(GuiPlugin):
    def __init__(self, gui):
        self.plugin_attrs = {}
        self.gui = gui

        on_hook("curses_start", self.do_cmds)

    def do_cmds(self):
        self.gui.callbacks["set_var"]("quiet", True)
        for cmd in cmds:
            self.gui.issue_cmd(cmd)
        self.gui.callbacks["set_var"]("quiet", False)



# Autocmd Plugin
# by Jack Miller
# v1.0

# This plugin allows you to automatically do commands on canto-curses startup
# Useful for machine-specific or environment-specific configuration

# Commands should be added to this list.
# They should be single commands (i.e. no &)

cmds = []

# Examples

# Setting browser / browser_txt based on TERM

import os

if "DISPLAY" not in os.environ:
    cmds.append("set browser elinks")
    cmds.append("set browser.text True")
else:
    cmds.append("set browser firefox")
    cmds.append("set browser.text False")

# The actual plugin workings below.

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_curses.gui import GuiPlugin
from canto_next.hooks import on_hook

class AutoCmdGui(GuiPlugin):
    def __init__(self, gui):
        self.plugin_attrs = {}
        self.gui = gui

        on_hook("curses_start", self.do_cmds)

    def do_cmds(self):
        self.gui.callbacks["set_var"]("quiet", True)
        for cmd in cmds:
            self.gui.issue_cmd(cmd)
        self.gui.callbacks["set_var"]("quiet", False)



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.plugins import Plugin, PluginHandler
from canto_next.hooks import on_hook, unhook_all

from .theme import FakePad, WrapPad, theme_print, theme_len, theme_reset, theme_border, prep_for_display
from .tagcore import tag_updater
from .config import story_needed_attrs
from .color import cc

import traceback
import logging
import curses

log = logging.getLogger("STORY")

class StoryPlugin(Plugin):
    pass

# The Story class is the basic wrapper for an item to be displayed. It manages
# its own state only because it affects its representation, it's up to a higher
# class to actually communicate state changes to the backend.

class Story(PluginHandler):
    def __init__(self, tag, id, callbacks):
        PluginHandler.__init__(self)

        self.callbacks = callbacks

        self.parent_tag = tag
        self.is_tag = False
        self.is_dead = False
        self.id = id
        self.pad = None

        self.selected = False
        self.marked = False

        # Are there changes pending?
        self.changed = True

        self.fresh_state = False
        self.fresh_tags = False

        self.width = 0

        # This is used by the rendering code.
        self.extra_lines = 0

        # Pre and post formats, to be used by plugins
        self.pre_format = ""
        self.post_format = ""

        # Offset globally and in-tag.
        self.offset = 0
        self.rel_offset = 0
        self.enumerated = False
        self.rel_enumerated = False

        # This should exist before the hook is setup, or the hook will fail.
        self.content = {}

        on_hook("curses_opt_change", self.on_opt_change, self)
        on_hook("curses_tag_opt_change", self.on_tag_opt_change, self)
        on_hook("curses_attributes", self.on_attributes, self)

        # Grab initial content, if any, the rest will be handled by the
        # attributes hook

        self.content = tag_updater.get_attributes(self.id)
        self.new_content = None

        self.plugin_class = StoryPlugin
        self.update_plugin_lookups()

    def die(self):
        self.is_dead = True
        unhook_all(self)

    def __eq__(self, other):
        if not other:
            return False
        if not hasattr(other, "id"):
            return False
        return self.id == other.id

    def __str__(self):
        return "story: %s" % self.id

    # On_attributes updates new_content. We don't lock because we don't
    # particularly care what version of new_content the next sync() call gets.

    def on_attributes(self, attributes):
        if self.id in attributes:
            new_content = attributes[self.id]

            if not (new_content is self.content):
                self.new_content = new_content

    def sync(self):
        if self.new_content == None:
            return

        old_content = self.content
        self.content = self.new_content
        self.new_content = None

        if 'canto-state' in old_content and self.fresh_state:
            self.content['canto-state'] = old_content['canto-state']
            self.fresh_state = False

        if 'canto-tags' in old_content and self.fresh_tags:
            self.content['canto-tags'] = old_content['canto-tags']
            self.fresh_tags = False

        self.need_redraw()

    def on_opt_change(self, config):
        if "taglist" in config and "border" in config["taglist"]:
            self.need_redraw()

        if "color" in config or "style" in config:
            self.need_redraw()

        if "story" not in config:
            return

        if "format_attrs" in config["story"]:
            needed_attrs = []
            for attr in config["story"]["format_attrs"]:
                if attr not in self.content:
                    needed_attrs.append(attr)
            if needed_attrs:
                log.debug("%s needs: %s", self.id, needed_attrs)
                tag_updater.need_attributes(self.id, needed_attrs)

        # All other story options are formats / enumerations, redraw.

        self.need_redraw()

    def on_tag_opt_change(self, config):
        tagname = self.callbacks["get_tag_name"]()
        if tagname in config and "enumerated" in config[tagname]:
            self.need_redraw()

    # Add / remove state. Return True if an actual change, False otherwise.

    def _handle_key(self, attr, key):
        if key not in self.content or self.content[key] == "":
            self.content[key] = []

        # Negative attribute
        if attr[0] == "-":
            attr = attr[1:]
            if attr == "marked":
                return self.unmark()
            elif attr in self.content[key]:
                self.content[key].remove(attr)
                self.need_redraw()
                return True

        # Toggle attribute
        elif attr[0] == '%':
            attr = attr[1:]
            if attr == "marked":
                if self.marked:
                    self.unmark()
                else:
                    self.mark()
            else:
                if attr in self.content[key]:
                    self.content[key].remove(attr)
                else:
                    self.content[key].append(attr)
                self.need_redraw()
                return True

        # Positive attribute
        else:
            if attr == "marked":
                return self.mark()
            elif attr not in self.content[key]:
                self.content[key].append(attr)
                self.need_redraw()
                return True
        return False

    # Simple wrapper to call tag's item_state_change callback on an actual
    # change.

    def handle_state(self, attr):
        r = self._handle_key(attr, "canto-state")
        if r:
            self.fresh_state = True
            self.callbacks["item_state_change"](self)
        return r

    def handle_tag(self, tag):
        r = self._handle_key(tag, "canto-tags")
        if r:
            self.fresh_tags = True
            self.callbacks["item_state_change"](self)
        return r

    def select(self):
        if not self.selected:
            self.selected = True
            self.need_redraw()

    def unselect(self):
        if self.selected:
            self.selected = False
            self.need_redraw()

    def mark(self):
        if not self.marked:
            self.marked = True
            self.need_redraw()
            return True
        return False

    def unmark(self):
        if self.marked:
            self.marked = False
            self.need_redraw()
            return True
        return False

    def set_offset(self, offset):
        if self.offset != offset:
            self.offset = offset
            self.need_redraw()

    def set_rel_offset(self, offset):
        if self.rel_offset != offset:
            self.rel_offset = offset
            self.need_redraw()

    # This is not useful in the interface,
    # so no redraw required on it.

    def set_sel_offset(self, offset):
        self.sel_offset = offset

    def need_redraw(self):
        self.changed = True
        self.callbacks["set_var"]("needs_redraw", True)

    def need_refresh(self):
        self.changed = True
        self.callbacks["set_var"]("needs_refresh", True)

    def eval(self):
        s = ""

        if "read" in self.content["canto-state"]:
            s += cc("read")
        else:
            s += cc("unread")

        if self.marked:
            s += cc("marked") + "[*]"

        if self.selected:
            s += cc("selected")

        s += prep_for_display(self.content["title"])

        if self.selected:
            s += cc.end("selected")

        if self.marked:
            s += cc.end("marked")

        if "read" in self.content["canto-state"]:
            s += cc.end("read")
        else:
            s += cc.end("unread")

        return s

    def lines(self, width):
        if width == self.width and not self.changed:
            return self.lns + self.extra_lines

        # Make sure we actually have all of the attributes needed
        # to complete the render.

        self.enumerated = self.callbacks["get_opt"]("story.enumerated")
        self.rel_enumerated = self.callbacks["get_tag_opt"]("enumerated")

        for attr in story_needed_attrs:
            if attr not in self.content:

                # Not having needed info is a good reason to
                # sync.

                self.sync()
                self.need_refresh()
                self.callbacks["release_gui"]()

                log.debug("%s still needs %s", self, attr)

                self.left = " "
                self.left_more = " "
                self.right = " "

                self.evald_string = "Waiting on content..."

                self.lns = 1
                return self.lns

        for attr in list(self.plugin_attrs.keys()):
            if not attr.startswith("edit_"):
                continue
            try:
                getattr(self, attr)()
            except:
                log.error("Error running story editing plugin")
                log.error(traceback.format_exc())

        self.evald_string = self.eval()

        taglist_conf = self.callbacks["get_opt"]("taglist")

        if taglist_conf["border"]:
            self.left = "%C%B" + theme_border("ls") + "%b %c"
            self.left_more = "%C%B" + theme_border("ls") + "%b     %c"
            self.right = "%C %B" + theme_border("rs") + "%b%c"
        else:
            self.left = "%C %c"
            self.left_more = "%C     %c"
            self.right = "%C %c"

        self.pad = None
        self.width = width
        self.changed = False

        self.lns = self.render(FakePad(width), width)
        return self.lns

    def pads(self, width):
        if self.pad and not self.changed:
            return self.lns

        self.pad = curses.newpad(self.lines(width), width)
        self.render(WrapPad(self.pad), width)
        return self.lns

    def render(self, pad, width):
        s = self.evald_string

        lines = 0

        try:
            while s:
                # Left border, for first line
                if lines == 0:
                    l = self.left

                # Left border, for subsequent lines (indent)
                else:
                    l = self.left_more

                s = theme_print(pad, s, width, l, self.right)

                # Handle overwriting with offset information

                if lines == 0:
                    header = ""
                    if self.enumerated:
                        header += cc("enum_hints") + "[" + str(self.offset) + "]%0"
                    if self.rel_enumerated:
                        header += cc("enum_hints") + "[" + str(self.rel_offset) + "]%0"
                    if header:
                        pad.move(0, 0)
                        theme_print(pad, header, width, "","", False, False)
                        try:
                            pad.move(1, 0)
                        except:
                            pass

                lines += 1

        # Render exceptions should be non-fatal. The worst
        # case scenario is that one story's worth of space
        # is going to be fucked up.

        except Exception as e:
            tb = traceback.format_exc()
            log.debug("Story exception:")
            log.debug("\n" + "".join(tb))

        # Reset theme counters
        theme_reset()

        return lines



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.plugins import Plugin
from canto_next.format import escsplit

from .tag import alltags
from .tagcore import tag_updater

from .locks import sync_lock
from .command import CommandHandler, cmd_execute, register_command, register_alias
from .text import ErrorBox, InfoBox
from .config import config
from .screen import Screen

from threading import Thread, Event
import traceback
import logging

log = logging.getLogger("GUI")

class GraphicalLog(logging.Handler):

    # We want to be able to catch logging output before the screen is actually
    # initialized in curses, and callbacks etc. are setup

    def __init__(self):
        logging.Handler.__init__(self)
        self.deferred_logs = []
        self.callbacks = None

        rootlog = logging.getLogger()
        rootlog.addHandler(self)

    def init(self, callbacks, screen):
        self.callbacks = callbacks
        self.screen = screen

    def _emit(self, var, window_type, record):
        if window_type not in self.screen.window_types:
            self.callbacks["set_var"](var, record.message)
            self.screen.add_window_callback(window_type)
        else:
            cur = self.callbacks["get_var"](var)
            cur += "\n" + record.message
            self.callbacks["set_var"](var, cur)

        self.callbacks["set_var"]("needs_refresh", True)

    def emit(self, record):

        # If we have no callbacks, GUI isn't initialized, assume that we only
        # want to have warns/errors displayed immediately on startup.

        if self.callbacks:
            quiet = self.callbacks["get_var"]("quiet")
        else:
            quiet = True
        if record.levelno == logging.INFO and quiet:
            return
        self.deferred_logs.append(record)

    # Call with sync_lock
    def flush_deferred_logs(self):
        for record in self.deferred_logs:
            if record.levelno in [ logging.INFO, logging.WARN ]:
                self._emit("info_msg", InfoBox, record)
            elif record.levelno == logging.ERROR:
                self._emit("error_msg", ErrorBox, record)
        self.deferred_logs = []

class GuiPlugin(Plugin):
    pass

class CantoCursesGui(CommandHandler):
    def __init__(self, backend, glog_handler):
        CommandHandler.__init__(self)
        self.plugin_class = GuiPlugin
        self.update_plugin_lookups()

        self.backend = backend
        self.winched = False

        self.update_interval = 0

        self.do_gui = Event()
        self.do_gui.set()

        self.working = False

        self.callbacks = {
            "set_var" : config.set_var,
            "get_var" : config.get_var,
            "set_conf" : config.set_conf,
            "get_conf" : config.get_conf,
            "set_tag_conf" : config.set_tag_conf,
            "get_tag_conf" : config.get_tag_conf,
            "set_defaults" : config.set_def_conf,
            "get_defaults" : config.get_def_conf,
            "set_feed_conf" : config.set_feed_conf,
            "get_feed_conf" : config.get_feed_conf,
            "get_opt" : config.get_opt,
            "set_opt" : config.set_opt,
            "get_tag_opt" : config.get_tag_opt,
            "set_tag_opt" : config.set_tag_opt,
            "release_gui" : self.release_gui,
            "force_sync" : self.force_sync,
            "switch_tags" : config.switch_tags,
        }

        log.debug("Starting curses.")

        self.alive = True
        self.sync_timer = 1
        self.sync_requested = True
        self.tags_to_sync = []

        self.screen = Screen(self.callbacks)
        self.screen.refresh()
        self.screen.redraw()

        self.glog_handler = glog_handler
        self.glog_handler.init(self.callbacks, self.screen)

        self.graphical_thread = Thread(target = self.run_gui)
        self.graphical_thread.daemon = True
        self.graphical_thread.start()

        register_command(self, "refresh", self.cmd_refresh, [], "Refetch everything from the daemon", "Base")
        register_command(self, "update", self.cmd_update, [], "Sync with daemon", "Base")
        register_command(self, "quit", self.cmd_quit, [], "Quit canto-curses", "Base")

        self.input_thread = Thread(target = self.run)
        self.input_thread.daemon = True
        self.input_thread.start()

    def force_sync(self):
        self.sync_requested = True
        self.sync_timer = 0
        self.release_gui()

    def release_gui(self):
        self.do_gui.set()

    def tick(self):
        c = self.callbacks["get_conf"]()
        if c["update"]["auto"]["enabled"]:
            self.sync_timer -= 1
            if self.sync_timer <= 0:
                self.sync_requested = True
                self.release_gui()
                self.sync_timer = c["update"]["auto"]["interval"]
        else:
            self.sync_timer = 1
            if self.sync_requested:
                self.release_gui()

    def winch(self):
        self.winched = True
        if not self.do_gui.is_set():
            self.release_gui()

    def cmd_refresh(self):
        # Will trigger a hook on completion that will cause refresh
        tag_updater.update()

    def cmd_update(self):
        self.force_sync()

    def cmd_quit(self):
        self.alive = False

    def cmdsplit(self, cmd):
        r = escsplit(cmd, " &")

        # lstrip all commands because we
        # want to use .startswith instead of a regex.
        return [ s.lstrip() for s in r ]

    def issue_cmd(self, cmd):
        sync_lock.acquire_write()
        try:
            r = cmd_execute(cmd)
            return r
        except Exception as e:
            log.error("Exception: %s" % e)
            log.error(traceback.format_exc())
        finally:
            sync_lock.release_write()

    def run(self):
        while self.alive:
            r = self.screen.get_key()

            # Get a list of all command handlers
            f = [self] + self.screen.get_focus_list()

            # We got a key, now resolve it to a command
            for win in reversed(f):
                cmd = win.key(r)
                if cmd:
                    break
            else:

                # Dismiss info box on any unbound key.

                if self.callbacks["get_var"]("info_msg"):
                    self.callbacks["set_var"]("info_msg", "")
                    self.callbacks["set_var"]("dispel_msg", False)
                    self.release_gui()
                continue

            cmds = self.cmdsplit(cmd)
            log.debug("Resolved to %s", cmds)

            # Now actually issue the commands

            for cmd in cmds:

                okay = False

                # Command is our one hardcoded command because it's special, and also shouldn't invoke itself.
                if cmd == "command":
                    subcmd = self.screen.input_callback(':')
                    log.debug("Got %s from user command", subcmd)
                    subcmds = self.cmdsplit(subcmd)
                    for subcmd in subcmds:
                        okay = self.issue_cmd(subcmd)
                        if not okay:
                            break
                else:
                    okay = self.issue_cmd(cmd)

                if not okay:
                    break

            # Let the GUI thread process, or realize it's dead.
            self.release_gui()

    def run_gui(self):
        while True:
            self.do_gui.wait()
            self.do_gui.clear()
            log.debug("gui thread released")

            if not self.alive:

                # Remove graphical log handler so log.infos don't screw up the
                # screen after it's dead.

                rootlog = logging.getLogger()
                rootlog.removeHandler(self.glog_handler)
                self.screen.exit()
                break

            sync_lock.acquire_write()

            self.glog_handler.flush_deferred_logs()

            partial_sync = False
            self.working = True

            if self.sync_requested:
                self.tags_to_sync = alltags[:]
                self.sync_requested = False
            else:
                for tag in alltags:
                    if (tag not in self.tags_to_sync) and (tag.tagcore.was_reset or\
                            (len(tag) == 0 and len(tag.tagcore) != 0)):
                        self.tags_to_sync.append(tag)

            if self.tags_to_sync:
                self.tags_to_sync[0].sync()
                self.tags_to_sync = self.tags_to_sync[1:]
                partial_sync = True

            needs_resize = self.callbacks["get_var"]("needs_resize") or self.winched
            needs_refresh = self.callbacks["get_var"]("needs_refresh")
            needs_redraw = self.callbacks["get_var"]("needs_redraw")

            self.callbacks["set_var"]("needs_resize", False)
            self.callbacks["set_var"]("needs_refresh", False)
            self.callbacks["set_var"]("needs_redraw", False)

            # Resize implies a refresh and redraw
            if needs_resize:
                self.winched = False
                self.screen.resize()
            else:
                if needs_refresh:
                    self.screen.refresh()

                if needs_redraw:
                    self.screen.redraw()

            needs_resize = self.callbacks["get_var"]("needs_resize") or self.winched
            needs_refresh = self.callbacks["get_var"]("needs_refresh")
            needs_redraw = self.callbacks["get_var"]("needs_redraw")

            # If we weren't able to clear the condition, then
            # we'll drop locks and immediately go again.

            if needs_resize or needs_refresh or needs_redraw or partial_sync:
                self.do_gui.set()
            else:
                self.working = False

            sync_lock.release_write()

    def get_opt_name(self):
        return "main"



# Xterm title set on selection change
# by Jack Miller
# v1.0

# Set to True if you want the selection title included.
USE_TITLE=False

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_next.hooks import on_hook
import locale
import os

prefcode = locale.getpreferredencoding()

def set_xterm_title(s):
    os.write(1, ("\033]0; %s \007" % s).encode(prefcode))

def clear_xterm_title():
    os.write(1, "\033]0; \007".encode(prefcode))

def xt_on_var_change(var_dict):
    if "selected" in var_dict:
        if var_dict["selected"] and "title" in var_dict["selected"].content:
            set_xterm_title("Canto - " + var_dict["selected"].content["title"])
        else:
            set_xterm_title("Canto")

if USE_TITLE:
    on_hook("curses_var_change", xt_on_var_change)
else:
    on_hook("curses_start", lambda: set_xterm_title("Canto"))

on_hook("curses_exit", clear_xterm_title)



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.hooks import call_hook, on_hook, unhook_all
from canto_next.plugins import Plugin, PluginHandler
from canto_next.rwlock import read_lock

from .locks import sync_lock, config_lock
from .theme import FakePad, WrapPad, theme_print, theme_reset, theme_border, prep_for_display
from .config import config
from .story import Story
from .color import cc

import traceback
import logging
import curses

log = logging.getLogger("TAG")

# TagCore provides the core tag functionality of keeping track of a list of IDs.

# The Tag class manages stories. Externally, it looks like a Tag takes IDs from
# the backend and renders an ncurses pad. No class other than Tag actually
# touches Story objects directly.

class TagPlugin(Plugin):
    pass

alltags = []

class Tag(PluginHandler, list):
    def __init__(self, tagcore, callbacks):
        list.__init__(self)
        PluginHandler.__init__(self)

        self.tagcore = tagcore
        self.tag = tagcore.tag
        self.is_tag = True
        self.updates_pending = 0

        self.pad = None
        self.footpad = None

        # Note that Tag() is only given the top-level CantoCursesGui
        # callbacks as it shouldn't be doing input / refreshing
        # itself.

        self.callbacks = callbacks.copy()

        # Modify our own callbacks so that *_tag_opt assumes
        # the current tag.

        self.callbacks["get_tag_opt"] =\
                lambda x : callbacks["get_tag_opt"](self.tag, x)
        self.callbacks["set_tag_opt"] =\
                lambda x, y : callbacks["set_tag_opt"](self.tag, x, y)
        self.callbacks["get_tag_name"] = lambda : self.tag

        # This could be implemented as a generic, top-level hook but then N
        # tags would have access to story objects they shouldn't have and
        # would have to check every items membership in self, which would be
        # pointless and time-consuming.

        self.callbacks["item_state_change"] =\
                self.on_item_state_change

        # Are there changes pending?
        self.changed = True

        self.selected = False
        self.marked = False

        # Information from last refresh
        self.footlines = 0
        self.extra_lines = 0
        self.width = 0

        self.collapsed = False
        self.border = False
        self.enumerated = False
        self.abs_enumerated = False

        # Formats for plugins to override
        self.pre_format = ""
        self.post_format = ""

        # Global indices (for enumeration)
        self.item_offset = -1
        self.visible_tag_offset = -1
        self.tag_offset = -1
        self.sel_offset = -1

        on_hook("curses_opt_change", self.on_opt_change, self)
        on_hook("curses_tag_opt_change", self.on_tag_opt_change, self)
        on_hook("curses_attributes", self.on_attributes, self)
        on_hook("curses_items_added", self.on_items_added, self)

        # Upon creation, this Tag adds itself to the
        # list of all tags.

        alltags.append(self)

        self.plugin_class = TagPlugin
        self.update_plugin_lookups()

    def die(self):
        log.debug("tag %s die()", self.tag)
        # Reset so items get die() called and everything
        # else is notified about items disappearing.

        for s in self:
            s.die()
        del self[:]

        alltags.remove(self)

        unhook_all(self)

    def on_item_state_change(self, item):
        self.need_redraw()

    def on_opt_change(self, opts):
        if "taglist" in opts and\
                ("tags_enumerated" in opts["taglist"] or\
                "tags_enumerated_absolute" in opts["taglist"] or\
                "border" in opts["taglist"]):
            self.need_redraw()

        if "tagobj" in opts:
            self.need_redraw()

        if "color" in opts or "style" in opts:
            self.need_redraw()

    def on_tag_opt_change(self, opts):
        if self.tag in list(opts.keys()):
            tc = opts[self.tag]
            if "collapsed" in tc:
                self.need_refresh()
            else:
                self.need_redraw()

    # Technically, we might want to hold sync_lock so that self[:] doesn't
    # change, but if we're syncing, the setting of needs_redraw isn't important
    # anymore, and if we're not, there's no issue.

    def on_attributes(self, attributes):
        for s in self:
            if s.id in attributes:
                self.need_redraw()
                break

    def on_items_added(self, tagcore, added):
        if tagcore == self.tagcore:
            cur_ids = self.get_ids()
            for story_id in added:
                if story_id not in cur_ids:
                    self.updates_pending += 1
            self.need_redraw()

    # We override eq so that empty tags don't evaluate
    # as equal and screw up things like enumeration.

    def __eq__(self, other):
        if other and (not other.is_tag or self.tag != other.tag):
            return False
        return list.__eq__(self, other)

    def __str__(self):
        return "%s" % self.tag[self.tag.index(':') + 1:]

    def get_id(self, id):
        for item in self:
            if item.id == id:
                return item

    def get_ids(self):
        return [ s.id for s in self ]

    # Inform the tag of global index of it's first item.
    def set_item_offset(self, offset):
        if self.item_offset != offset:
            self.item_offset = offset
            for i, item in enumerate(self):
                item.set_offset(offset + i)

    # Note that this cannot be short-cut (i.e.
    # copout if sel_offset is already equal)
    # because it's possible that it's the same
    # without the items having ever been updated.

    # Alternatively, we could reset them in
    # on_tag_opt_change, but since the sel
    # offset does not cause a redraw, there's
    # no point.

    def set_sel_offset(self, offset):
        self.sel_offset = offset

        if not self.callbacks["get_tag_opt"]("collapsed"):
            for i, item in enumerate(self):
                item.set_sel_offset(offset + i)

    def set_visible_tag_offset(self, offset):
        if self.visible_tag_offset != offset:
            self.visible_tag_offset = offset
            self.need_redraw()

    def set_tag_offset(self, offset):
        if self.tag_offset != offset:
            self.tag_offset = offset
            self.need_redraw()

    def select(self):
        if not self.selected:
            self.selected = True
            self.need_redraw()

    def unselect(self):
        if self.selected:
            self.selected = False
            self.need_redraw()

    def need_refresh(self):
        self.changed = True
        self.callbacks["set_var"]("needs_refresh", True)

    def need_redraw(self):
        self.changed = True
        self.callbacks["set_var"]("needs_redraw", True)

    def eval(self):
        # Make sure to strip out the category from category:name
        tag = self.tag.split(':', 1)[1]

        unread = len([s for s in self\
                if "canto-state" not in s.content or\
                "read" not in s.content["canto-state"]])

        s = ""
        if self.selected:
            s += cc("selected")

        if self.collapsed:
            s += "[+]"
        else:
            s += "[-]"

        s += " " + tag + " "

        s += "[" + cc("unread") + str(unread) + cc.end("unread") + "]"
        if self.updates_pending:
            s += " [" + cc("pending") + str(self.updates_pending) + cc.end("pending") + "]"

        if self.selected:
            s += cc.end("selected")

        return s

    def lines(self, width):
        if width == self.width and not self.changed:
            return self.lns

        taglist_conf = self.callbacks["get_opt"]("taglist")

        self.collapsed = self.callbacks["get_tag_opt"]("collapsed")
        self.border = taglist_conf["border"]
        self.enumerated = taglist_conf["tags_enumerated"]
        self.abs_enumerated = taglist_conf["tags_enumerated_absolute"]

        extra_tags = self.callbacks["get_tag_conf"](self.tag)['extra_tags']

        self.pad = None
        self.footpad = None
        self.width = width
        self.changed = False

        self.evald_string = self.eval()

        self.lns = self.render_header(width, FakePad(width))
        self.footlines = self.render_footer(width, FakePad(width))

        return self.lns

    def pads(self, width):
        if self.pad and (self.footpad or not self.footlines) and not self.changed:
            return self.lns

        self.pad = curses.newpad(self.lines(width), width)
        self.render_header(width, WrapPad(self.pad))

        if self.footlines:
            self.footpad = curses.newpad(self.footlines, width)
            self.render_footer(width, WrapPad(self.footpad))
        return self.lns

    def render_header(self, width, pad):
        s = self.evald_string
        lines = 0

        try:
            while s:
                s = theme_print(pad, s, width, "", "")

                if lines == 0:
                    header = ""
                    if self.enumerated:
                        header += cc("enum_hints") + "[" + str(self.visible_tag_offset) + "]%0"
                    if self.abs_enumerated:
                        header += cc("enum_hints") + "[" + str(self.tag_offset) + "]%0"
                    if header:
                        pad.move(0, 0)
                        theme_print(pad, header, width, "", "", False, False)
                        try:
                            pad.move(1, 0)
                        except:
                            pass
                lines += 1

            if not self.collapsed and self.border:
                theme_print(pad, theme_border("ts") * (width - 2), width,\
                        "%B"+ theme_border("tl"), theme_border("tr") + "%b")
                lines += 1
        except Exception as e:
            tb = traceback.format_exc()
            log.debug("Tag exception:")
            log.debug("\n" + "".join(tb))

        theme_reset()

        return lines

    def render_footer(self, width, pad):
        if not self.collapsed and self.border:
            theme_print(pad, theme_border("bs") * (width - 2), width,\
                    "%B" + theme_border("bl"), theme_border("br") + "%b")
            theme_reset()
            return 1
        return 0

    # Synchronize this Tag with its TagCore

    def sync(self, force=False):
        if force or self.tagcore.changes:
            sel = self.callbacks["get_var"]("selected")

            self.tagcore.lock.acquire_read()

            self.tagcore.ack_changes()

            # Get sorted ids, along with their (unsorted) positions in the
            # original lists.

            sorted_ids = [ (x, x.id, i) for (i, x) in enumerate(self) ]
            sorted_ids.sort(key=lambda x: x[1])

            tagcore_sorted_ids = list(enumerate(self.tagcore))
            tagcore_sorted_ids.sort(key=lambda x: x[1])

            new_ids = []
            current_stories = []
            old_stories = []

            for story, s_id, place in sorted_ids:
                while tagcore_sorted_ids and s_id > tagcore_sorted_ids[0][1]:
                    new_ids.append(tagcore_sorted_ids.pop(0))

                if not tagcore_sorted_ids or s_id < tagcore_sorted_ids[0][1]:
                    if sel and (not sel.is_tag) and (s_id == sel.id):

                        # If we preserve the selection in an "undead" state, then
                        # we keep set tagcore changed so that the next sync operation
                        # will re-evaluate it.

                        self.tagcore.changed()
                        place = -1
                        current_stories.append((place, story))
                    else:
                        old_stories.append(story)
                else:
                    place = tagcore_sorted_ids.pop(0)[0]
                    current_stories.append((place, story))

            # Grab any remaining new items
            new_ids += tagcore_sorted_ids

            self.tagcore.lock.release_read()

            new_stories = [ (p, Story(self, x, self.callbacks)) for (p, x) in new_ids ]

            call_hook("curses_stories_added", [ self, [ x for (p, x) in new_stories ]])

            del self[:]

            conf = config.get_conf()
            if conf["update"]["style"] == "maintain" or self.tagcore.was_reset:
                self.tagcore.was_reset = False
                current_stories += new_stories
                current_stories.sort()
                self.extend([ x[1] for x in current_stories ])
            else:
                current_stories.sort()
                new_stories.sort()
                if conf["update"]["style"] == "append":
                    current_stories += new_stories
                    self.extend([ x[1] for x in current_stories ])
                else:
                    new_stories += current_stories
                    self.extend([ x[1] for x in new_stories ])

            for story in old_stories:
                story.die()

            # Properly dispose of the remaining stories

            call_hook("curses_stories_removed", [ self, old_stories ])

            # Trigger a refresh so that classes above (i.e. TagList) will remap
            # items

            self.need_refresh()

        # Pass the sync onto story objects
        for s in self:
            s.sync()

        self.updates_pending = 0



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.hooks import on_hook, unhook_all

from .theme import FakePad, WrapPad, theme_print, theme_lstrip, theme_border, theme_reset
from .command import register_commands, unregister_command
from .guibase import GuiBase
from .color import cc

import logging
import curses

log = logging.getLogger("TEXTBOX")

class TextBox(GuiBase):
    def init(self, pad, callbacks, lstrip=True):
        GuiBase.init(self)

        self.pad = pad

        self.max_offset = 0

        self.callbacks = callbacks

        self.lstrip = lstrip
        self.text = ""

        # This relies on the actual subclasses (i.e. Reader) to cleanup.

        cmds = { "page-down" : (self.cmd_page_down, [], "Next page of text"),
                "page-up" : (self.cmd_page_up, [], "Previous page of text"),
                "scroll-up" : (self.cmd_scroll_up, [], "Scroll up"),
                "scroll-down" : (self.cmd_scroll_down, [], "Scroll down"),
        }

        register_commands(self, cmds)

    def get_offset(self):
        return self.callbacks["get_var"](self.get_opt_name() + "_offset")

    def set_offset(self, offset):
        self.callbacks["set_var"](self.get_opt_name() + "_offset", offset)

    def update_text(self):
        pass

    def refresh(self):
        self.height, self.width = self.pad.getmaxyx()

        fp = FakePad(self.width)
        lines = self.render(fp)

        # Create pre-rendered pad
        self.fullpad = curses.newpad(lines, self.width)
        self.render(WrapPad(self.fullpad))

        # Update offset based on new display properties.
        self.max_offset = max((lines - 1) - (self.height - 1), 0)

        offset = min(self.get_offset(), self.max_offset)
        self.set_offset(offset)
        self.callbacks["set_var"]("needs_redraw", True)

    def redraw(self):
        offset = self.get_offset()
        tb, lb, bb, rb = self.callbacks["border"]()

        # Overwrite visible pad with relevant area of pre-rendered pad.
        self.pad.erase()

        realheight = min(self.height, self.fullpad.getmaxyx()[0]) - 1

        top = 0
        if tb:
            self.pad.move(0, 0)
            self.render_top_border(WrapPad(self.pad))
            top += 1

        self.fullpad.overwrite(self.pad, offset, 0, top, 0,\
                realheight - top, self.width - 1)

        if bb:
            # If we're not floating, then the bottom border
            # belongs at the bottom of the given window.

            if not self.callbacks["floating"]():
                padheight = self.pad.getmaxyx()[0] -1
                self.pad.move(padheight - 1, 0)
                self.render_bottom_border(WrapPad(self.pad))
                self.pad.move(padheight - 1, 0)
            else:
                self.pad.move(realheight - 1, 0)
                self.render_bottom_border(WrapPad(self.pad))
                self.pad.move(realheight - 1, 0)
        else:
            self.pad.move(realheight - 1, 0)

        self.callbacks["refresh"]()

    def render_top_border(self, pad):
        tb, lb, bb, rb = self.callbacks["border"]()

        lc = " "
        if lb:
            lc = "%C" + theme_border("tl") + "%c"

        rc = " "
        if rb:
            rc = "%C" + theme_border("tr") + "%c"

        mainbar = "%C" + (theme_border("ts") * (self.width - 1)) + "%c"
        theme_print(pad, mainbar, self.width, lc, rc)

    def render_bottom_border(self, pad):
        tb, lb, bb, rb = self.callbacks["border"]()

        lc = " "
        if lb:
            lc = "%C" + theme_border("bl") + "%c"

        rc = " "
        if rb:
            rc = "%C" + theme_border("br") + "%c"

        mainbar = "%C" + (theme_border("ts") * (self.width - 1)) + "%c"
        theme_print(pad, mainbar, self.width, lc, rc)

    def render(self, pad):
        self.update_text()

        tb, lb, bb, rb = self.callbacks["border"]()
        s = self.text

        lines = 0

        # Account for potential top border rendered on redraw.
        if tb:
            lines += 1

        # Prepare left and right borders

        l = " "
        if lb:
            l = "%C" + theme_border("ls") + " %c"
        r = " "
        if rb:
            r = "%C " + theme_border("rs") + "%c"

        # Render main content

        while s:
            if self.lstrip:
                s = theme_lstrip(pad, s)
            if s:
                s = theme_print(pad, s, self.width, l, r)
                lines += 1

        # Account for potential bottom rendered on redraw.
        if bb:
            lines += 1

        theme_reset()

        # Return one extra line because the rest of the reader
        # code knows to avoid the dead cell on the bottom right
        # of every curses pad.

        return lines + 1

    def cmd_scroll_up(self):
        self._relscroll(-1)

    def cmd_scroll_down(self):
        self._relscroll(1)

    def cmd_page_up(self):
        self._relscroll(-1 * (self.height - 1))

    def cmd_page_down(self):
        self._relscroll(self.height - 1)

    def _relscroll(self, factor):
        offset = self.get_offset()
        offset += factor
        offset = min(offset, self.max_offset)
        offset = max(offset, 0)

        self.set_offset(offset)
        self.callbacks["set_var"]("needs_redraw", True)

    def is_input(self):
        return False

    def get_opt_name(self):
        return "textbox"

    def get_height(self, mheight):
        return mheight

    def get_width(self, mwidth):
        return mwidth

class VarBox(TextBox):
    def init(self, pad, callbacks, var):
        TextBox.init(self, pad, callbacks, False)
        unregister_command(self, "bind")
        self.var = var
        self.value = self.callbacks["get_var"](var)

        on_hook("curses_var_change", self.on_var_change, self)

    def on_var_change(self, change):
        if self.var in change:
            self.value = change[self.var]
            if self.value == "":
                self.cmd_destroy()
            self.callbacks["set_var"]("needs_refresh", True)

    def cmd_destroy(self):
        unhook_all(self)
        TextBox.cmd_destroy(self)

class InfoBox(VarBox):
    def init(self, pad, callbacks):
        VarBox.init(self, pad, callbacks, "info_msg")

    def update_text(self):
        self.text = self.value

    def get_opt_name(self):
        return "infobox"

class ErrorBox(VarBox):
    def init(self, pad, callbacks):
        VarBox.init(self, pad, callbacks, "error_msg")

    def update_text(self):
        self.text = cc("error") + self.value + "%0"

    def get_opt_name(self):
        return "errorbox"



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.rwlock import RWLock
from canto_next.hooks import call_hook, on_hook

from .subthread import SubThread
from .locks import config_lock
from .config import config, story_needed_attrs

import traceback
import logging

log = logging.getLogger("TAGCORE")

alltagcores = []

class TagCore(list):
    def __init__(self, tag):
        list.__init__(self)
        self.tag = tag

        self.changes = False
        self.was_reset = False

        self.lock = RWLock("lock: %s" % tag)
        alltagcores.append(self)

    # change functions must be called holding lock

    def ack_changes(self):
        self.changes = False

    def changed(self):
        self.changes = True

    def set_items(self, ids):
        self.lock.acquire_write()

        del self[:]
        self.extend(ids)
        self.changed()

        self.lock.release_write()

class TagUpdater(SubThread):
    def init(self, backend):
        SubThread.init(self, backend)

        self.updating = []

        self.attributes = {}
        self.lock = RWLock("tagupdater")

        self.start_pthread()

        # Setup automatic attributes.

        # We know we're going to want at least these attributes for
        # all stories, as they're part of the fallback format string.

        self.needed_attrs = [ "title", "canto-state", "canto-tags", "link", "enclosures" ]

        tsa = config.get_opt("taglist.search_attributes")

        for attrlist in [ story_needed_attrs, tsa ]:
            for sa in attrlist:
                if sa not in self.needed_attrs:
                    self.needed_attrs.append(sa)

        self.write("AUTOATTR", self.needed_attrs)

        # Lock config_lock so that strtags doesn't change and we miss
        # tags.

        config_lock.acquire_read()

        strtags = config.get_var("strtags")

        # Request initial information, instantiate TagCores()

        self.write("WATCHTAGS", strtags)
        for tag in strtags:
            self.on_new_tag(tag)

        on_hook("curses_new_tag", self.on_new_tag)
        on_hook("curses_del_tag", self.on_del_tag)
        on_hook("curses_stories_removed", self.on_stories_removed)
        on_hook("curses_def_opt_change", self.on_def_opt_change)

        config_lock.release_read()

    def on_new_tag(self, tag):
        self.write("WATCHTAGS", [ tag ])
        self.prot_tagchange(tag)
        call_hook("curses_new_tagcore", [ TagCore(tag) ])

    def on_del_tag(self, tag):
        for tagcore in alltagcores[:]:
            if tagcore.tag == tag:
                if len(tagcore):
                    call_hook("curses_items_removed", [ tagcore, tagcore ] )
                    tagcore.set_items([])
                call_hook("curses_del_tagcore", [ tagcore ])
                alltagcores.remove(tagcore)
                while tagcore in self.updating:
                    self.updating.remove(tagcore)
                return

    # Once they've been removed from the GUI, their attributes can be forgotten
    def on_stories_removed(self, tag, items):
        tagcore = None
        for tc in alltagcores:
            if tc.tag == tag.tag:
                tagcore = tc
                break
        else:
            log.warn("Couldn't find tagcore for removed story tag %s" % tag.tag)

        self.lock.acquire_write()
        for item in items:
            if tagcore and item.id in tc:
                log.debug("%s still in tagcore, not removing", item.id)
                continue
            if item.id in self.attributes:
                del self.attributes[item.id]
        self.lock.release_write()

    # Changes to global filters should force a full refresh.

    def on_def_opt_change(self, defaults):
        if 'global_transform' in defaults:
            log.debug("global_transform changed, forcing reset + update")
            self.update()

    def prot_attributes(self, d):
        # Update attributes, and then notify everyone to grab new content.
        self.lock.acquire_write()

        for key in d.keys():
            if key in self.attributes:

                # If we're updating, we want to create a whole new dict object
                # so that our stories dicts don't get updated without a sync

                cp = self.attributes[key].copy()
                cp.update(d[key])
                self.attributes[key] = cp
            else:
                self.attributes[key] = d[key]
        self.lock.release_write()

        call_hook("curses_attributes", [ self.attributes ])

    def prot_items(self, updates):
        # Daemon should now only return with one tag in an items response

        tag = list(updates.keys())[0]

        for have_tag in alltagcores:
            if have_tag.tag == tag:
                break
        else:
            return

        sorted_updated_ids = list(enumerate(updates[tag]))
        sorted_updated_ids.sort(key=lambda x : x[1])

        sorted_current_ids = list(enumerate(have_tag))
        sorted_current_ids.sort(key=lambda x : x[1])

        new_ids = []
        cur_ids = []
        old_ids = []

        for c_place, c_id in sorted_current_ids:
            while sorted_updated_ids and c_id > sorted_updated_ids[0][1]:
                new_ids.append(sorted_updated_ids.pop(0))

            if not sorted_updated_ids or c_id < sorted_updated_ids[0][1]:
                old_ids.append(c_id)
            else:
                cur_ids.append(sorted_updated_ids.pop(0))

        new_ids += sorted_updated_ids

        all_ids = new_ids + cur_ids
        all_ids.sort()

        have_tag.set_items([x[1] for x in all_ids])

        if new_ids:
            call_hook("curses_items_added", [ have_tag, [x[1] for x in new_ids] ] )

        if old_ids:
            call_hook("curses_items_removed", [ have_tag, old_ids ] )

        if have_tag in self.updating:
            have_tag.was_reset = True
            call_hook("curses_tag_updated", [ have_tag ])
            self.updating.remove(have_tag)
            if self.updating == []:
                call_hook("curses_update_complete", [])

    def prot_itemsdone(self, tag):
        pass

    def prot_tagchange(self, tag):
        self.write("ITEMS", [ tag ])

    # The following is the external interface to tagupdater.

    def update(self):
        self.reset()
        strtags = config.get_var("strtags")
        for tag in strtags:
            self.write("ITEMS", [ tag ])

    def reset(self):
        self.updating += alltagcores
        return True

    def transform(self, name, transform):
        self.write("TRANSFORM", { name : transform })
        self.reset()

    # Writes are already serialized, so in the meantime, we protect
    # self.attributes and self.needed_attrs with our lock.

    def get_attributes(self, id):
        r = {}
        self.lock.acquire_read()
        if id in self.attributes:
            r = self.attributes[id]
        self.lock.release_read()
        return r

    # This takes a fat argument because callers need to be able to curry
    # together multiple sets so stuff like 'item-state read *' don't generate
    # thousands of SETATTRIBUTES calls and take forever

    def set_attributes(self, arg):
        self.lock.acquire_write()
        self.write("SETATTRIBUTES", arg)
        self.lock.release_write()

    def request_attributes(self, id, attrs):
        self.write("ATTRIBUTES", { id : attrs })

    def need_attributes(self, id, attrs):
        self.lock.acquire_write()

        needed = self.needed_attrs[:]
        updated = False

        for attr in attrs:
            if attr not in needed:
                needed.append(attr)
                updated = True

        if updated:
            self.needed_attrs = needed
            self.write("AUTOATTR", self.needed_attrs)

        self.lock.release_write()

        # Even if we didn't update this time, make sure we attempt to get this
        # id's new needed attributes.

        self.write("ATTRIBUTES", { id : needed })

tag_updater = TagUpdater()



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.hooks import on_hook
from canto_next.plugins import Plugin
from canto_next.remote import assign_to_dict, access_dict

from .command import CommandHandler, register_commands, register_arg_types, unregister_all, _string, register_aliases, commands, command_help, groups
from .tagcore import tag_updater
from .theme import prep_for_display
from .config import needs_eval, config

import logging

log = logging.getLogger("COMMON")

import subprocess
import tempfile
import urllib.request, urllib.error, urllib.parse
import shlex
import sys

import os
import os.path

class BasePlugin(Plugin):
    pass

class GuiBase(CommandHandler):
    def init(self):
        args = {
            "key": ("[key]: Simple keys (a), basic chords (C-r, M-a), or named whitespace like space or tab", _string),
            "command": ("[command]: Any canto-curses command. (Will show current binding if not given)\n  Simple: goto\n  Chained: foritems \\\\& goto \\\\& item-state read \\\\& clearitems \\\\& next-item", self.type_unescape_command),
            "remote-cmd": ("[remote cmd]", self.type_remote_cmd),
            "url" : ("[URL]", _string),
            "help-command" : ("[help-command]: Any canto-curses command, if blank, 'any' or unknown, will display help overview", self.type_help_cmd),
            "config-option" : ("[config-option]: Any canto-curses option", self.type_config_option),
            "config-section" : ("[config-section]: A canto-curses config section", self.type_config_section),
            "executable" : ("[executable]: A program in your PATH", self.type_executable),
        }

        cmds = {
            "bind" : (self.cmd_bind, [ "key", "command" ], "Add or query %s keybinds" % self.get_opt_name()),
            "transform" : (self.cmd_transform, ["string"], "Set user transform"),
            "remote addfeed" : (lambda x : self.cmd_remote("addfeed", x), ["url"], "Subscribe to a feed"),
            "remote listfeeds" : (lambda : self.cmd_remote("listfeeds", ""), [], "List feeds"),
            "remote": (self.cmd_remote, ["remote-cmd", "string"], "Give a command to canto-remote"),
            "destroy": (self.cmd_destroy, [], "Destroy this %s" % self.get_opt_name()),
            "set" : (self.cmd_set, ["config-option", "string"],

"""Set configuration options

Common options:
    %BSetting browser options%b

    :set browser.path /path/to/browser
    :set browser.text [True|False]
        - True: text browser like elinks
        * False: graphical browser like Firefox

    %BSetting update styles%b

    :set update.style [maintain|append|prepend]
        - maintain: re-sort items into feed
        * append: add new items to the end of feeds
        - prepend: add new items to the top of feeds

    :set update.auto.interval <seconds>
    :set update.auto.enabled [True|False]
        - True: interface will automatically add new items
        * False: new items have to be requested with :update (\\\ by default)

    %BTaglist cursor settings%b

    :set taglist.cursor.type [edge|top|middle|bottom]
        * edge: list scrolls near the edge of the screen
        - top: cursor at top of screen
        - middle: cursor at middle of the screen
        - bottom: cursor at bottom of the screen

    :set taglist.cursor.edge <lines>
        - How close to the actual edge of the screen the cursor
            gets before scrolling (defaults to 5)

    :set taglist.cursor.scroll [scroll|page]
        * scroll: standard scrolling
        - page: page through items

    %BChanging feed defaults%b

    :set defaults.keep_time <seconds>
        - How long items are kept after they disappear from source data
    :set defaults.keep_unread [True|False]
        - True: unread items will be keep forever
        * False: unread items will be discarded if old enough

    %BChanging a setting per-feed (with item selected)%b

    :set feed.keep_time
    :set feed.keep_unread

    %BChanging filters/sorts%b

    :set defaults.global_transform <transform>
        - Basic transforms
            - None (see all items all the time)
            * filter_read (filter out read items)

    :set tag.transform <transform>
"""),
            "set browser.path" : (lambda x : self.cmd_set("browser.path", x), ["executable"], "Set desired browser"),
            "reset-config" : (self.cmd_reset_config, [ "config-section" ], "Reset canto-curses config (won't touch daemon / feed settings)")
        }

        help_cmds = {
            "help" : (self.cmd_help, ["help-command"], "Get help on a specific command")
        }

        aliases = {
            "add" : "remote addfeed",
            "del" : "remote delfeed",
            "list" : "remote listfeeds",

            # Compatibility / evaluation aliases
            "set global_transform" : "set defaults.global_transform",
            "set keep_time" : "set defaults.keep_time",
            "set keep_unread" : "set defaults.keep_unread",
            "set browser " : "set browser.path ",
            "set txt_browser " : "set browser.text ",
            "set update.auto " : "set update.auto.enabled ",
            "set border" : "set taglist.border",

            "filter" : "transform",
            "sort" : "transform",

            "next-item" : "rel-set-cursor 1",
            "prev-item" : "rel-set-cursor -1",
        }

        register_arg_types(self, args)

        register_commands(self, cmds, "Base")
        register_commands(self, help_cmds, "Help")

        register_aliases(self, aliases)

        self.editor = None

        self.plugin_class = BasePlugin
        self.update_plugin_lookups()

    def cmd_destroy(self):
        self.callbacks["die"](self)

    def die(self):
        unregister_all(self)

    # Provide completions, but we don't care to verify settings.

    def type_executable(self):
        executables = []
        for path_dir in os.environ["PATH"].split(os.pathsep):
            try:
                for f in os.listdir(path_dir):
                    fullpath = os.path.join(path_dir, f)
                    if os.path.isfile(fullpath) and os.access(fullpath, os.X_OK):
                        executables.append(f)

            # PATH directories aren't guaranteed to exist and a myriad of other
            # errors should just silently move on. Worst case is incomplete
            # list of completions.

            except:
                pass

        return (executables, lambda x : (True, x))

    def _fork(self, path, href, text):
        pid = os.fork()

        # Parents can now bail.
        if pid:
            return pid

        # Make sure that we quote href such that malicious URLs like
        # "http://example.com & rm -rf ~/" won't be interpreted by the shell.

        href = shlex.quote(href)

        # A lot of programs don't appreciate
        # having their fds closed, so instead
        # we dup them to /dev/null.

        fd = os.open("/dev/null", os.O_RDWR)
        os.dup2(fd, sys.stderr.fileno())

        if not text:
            os.setpgid(os.getpid(), os.getpid())
            os.dup2(fd, sys.stdout.fileno())
            os.dup2(fd, sys.stdin.fileno())

        if "%u" in path:
            path = path.replace("%u", href)
        elif href:
            path = path + " " + href

        os.execv("/bin/sh", ["/bin/sh", "-c", path])

        # Just in case.
        sys.exit(0)

    def type_remote_cmd(self):
        remote_cmds = [ "help", "addfeed", "listfeeds", "delfeed",
                "force-update", "config", "one-config", "export",
                "import", "kill" ]
        return (remote_cmds, lambda x : (True, x))

    def _remote_argv(self, argv):
        loc_args = self.callbacks["get_var"]("location")
        argv = [argv[0]] + loc_args + argv[1:]

        log.debug("Calling remote: %s", argv)

        # check_output return bytes, we must decode.
        out = subprocess.check_output(argv).decode()

        if out:
            log.debug("Output:")
            log.debug(out.rstrip())

            # Strip anything that could be misconstrued as style
            # from remote output.

            out = out.replace("%","\\%")

            log.info(out.rstrip())

    def _remote(self, args):
        args = "canto-remote " + args

        # Add location args, so the remote is connecting
        # to the correct daemon.

        self._remote_argv(shlex.split(args))

    def remote_args(self, args):
        return self.string(args, "remote: ")

    def cmd_remote(self, remote_cmd, args):
        self._remote("%s %s" % (remote_cmd, args))

    def _goto(self, urls):
        browser = self.callbacks["get_conf"]()["browser"]

        if not browser["path"]:
            log.error("No browser defined! Cannot goto.")
            return

        if browser["text"]:
            self.callbacks["pause_interface"]()

        for url in urls:
            pid = self._fork(browser["path"], url, browser["text"])
            if browser["text"]:
                os.waitpid(pid, 0)

        if browser["text"]:
            self.callbacks["unpause_interface"]()

    def cmd_transform(self, transform):
        tag_updater.transform("user", transform)
        tag_updater.update()

    def type_unescape_command(self):
        def validate_uescape_command(x):
            # Change the escaped '&' from shlex into a raw &
            return (True, x.replace(" '&' ", " & "))
        return (None, validate_uescape_command)

    def cmd_bind(self, key, cmd):
        self.bind(key, cmd.lstrip().rstrip(), True)

    def bind(self, key, cmd, overwrite=False):
        opt = self.get_opt_name()
        key = self.translate_key(key)
        c = self.callbacks["get_conf"]()
        if not cmd:
            if key in c[opt]["key"]:
                log.info("[%s] %s = %s" % (opt, key, c[opt]["key"][key]))
                return True
            else:
                return False
        else:
            if key in c[opt]["key"] and c[opt]["key"][key] and not overwrite:
                log.debug("%s already bound to %s", key, c[opt]["key"][key])
                return False

            log.debug("Binding %s.%s to %s", opt, key, cmd)

            c[opt]["key"][key] = cmd
            self.callbacks["set_conf"](c)
            return True

    def type_help_cmd(self):
        help_cmds = commands()

        def help_validator(x):
            if x in ["commands", "cmds"]:
                return (True, 'commands')
            if x in help_cmds:
                return (True, x)
            return (True, 'all')

        return (help_cmds, help_validator)

    def cmd_help(self, cmd):
        if self.callbacks["get_var"]("info_msg"):
            self.callbacks["set_var"]("info_msg", "")

        if cmd == 'all':
            log.info("%BHELP%b\n")
            log.info("This is a list of available keybinds.\n")
            log.info("For a list of commands, type ':help commands'\n")
            log.info("For help with a specific command, type ':help [command]'\n")
            log.info("%BBinds%b")

            config = self.callbacks["get_conf"]()

            for optname in [ "main", "infobox", "taglist", "reader" ]:
                if "key" in config[optname] and list(config[optname]["key"].keys()) != []:
                    maxbindl = max([ len(x) for x in config[optname]["key"].keys() ]) + 1
                    log.info("\n%B" + optname.title() + "%b\n")
                    for bind in sorted(config[optname]["key"]):
                        bindeff = prep_for_display(bind + (" " * (maxbindl - len(bind))))
                        cmd = prep_for_display(config[optname]["key"][bind])
                        log.info("%s %s" % (bindeff, cmd))

        elif cmd == 'commands':
            for group in sorted(groups()):
                log.info("%B" + group + "%b\n")
                tmp = {}
                for c in commands(group):
                    tmp[c] = command_help(c)

                maxcmdl = max([ len(x) for x in tmp ])
                for c in sorted(tmp.keys()):
                    ceff = c + (" " * (maxcmdl - len(c)))
                    log.info("%s - %s" % (ceff, tmp[c]))
                log.info("")
        else:
            log.info(command_help(cmd, True))

    # Validate a single config option
    # Will offer completions for any recognized config option
    # Will *not* reject validly formatted options that don't already exist

    def _get_current_config_options(self, obj, stack):
        r = []

        for item in obj.keys():
            stack.append(item)

            if type(obj[item]) == dict:
                r.extend(self._get_current_config_options(obj[item], stack[:]))
            else:
                r.append(shlex.quote(".".join(stack)))

            stack = stack[:-1]

        return r

    def type_config_option(self):
        conf = self.callbacks["get_conf"]()

        possibles = self._get_current_config_options(conf, [])
        possibles.extend(self._get_current_config_options(config.daemon_defaults, ["defaults"]))
        possibles.extend(self._get_current_config_options(config.tag_template_config, ["tag"]))
        possibles.extend(self._get_current_config_options({ "rate" : 10, "keep_time" : 86400, "keep_unread" : False}, ["feed"]))
        possibles.sort()

        return (possibles, lambda x : (True, x))

    def type_config_section(self):
        conf = self.callbacks["get_conf"]()
        possibles = [ "" ] + list(conf.keys())
        return (possibles, lambda x : (x in possibles, x))

    def cmd_set(self, opt, val):
        log.debug("SET: %s '%s'", opt, val)

        evaluate = needs_eval(opt)

        if val != "" and evaluate:
            log.debug("Evaluating...")
            try:
                val = eval(val)
            except Exception as e:
                log.error("Couldn't eval '%s': %s" % (val, e))
                return

        if opt.startswith("defaults."):
            conf = { "defaults" : self.callbacks["get_defaults"]() }

            if val != "":
                assign_to_dict(conf, opt, val)
                self.callbacks["set_defaults"](conf["defaults"])
        elif opt.startswith("feed."):
            sel = self.callbacks["get_var"]("selected")
            if not sel:
                log.info("Feed settings only work with a selected item")
                return

            if sel.is_tag:
                try_tag = sel
            else:
                try_tag = sel.parent_tag

            if not try_tag.tag.startswith("maintag:"):
                log.info("Selection is in a user tag, cannot set feed settings")
                return

            name = try_tag.tag[8:]

            conf = { "feed" : self.callbacks["get_feed_conf"](name) }

            if val != "":
                assign_to_dict(conf, opt, val)
                self.callbacks["set_feed_conf"](name, conf["feed"])
        elif opt.startswith("tag."):
            sel = self.callbacks["get_var"]("selected")
            if not sel:
                log.info("Tag settings only work with a selected item")
                return

            if sel.is_tag:
                tag = sel
            else:
                tag = sel.parent_tag

            conf = { "tag" : self.callbacks["get_tag_conf"](tag.tag) }

            if val != "":
                assign_to_dict(conf, opt, val)
                self.callbacks["set_tag_conf"](tag.tag, conf["tag"])
        else:
            conf = self.callbacks["get_conf"]()

            if val != "":
                assign_to_dict(conf, opt, val)
                self.callbacks["set_conf"](conf)

        ok, val = access_dict(conf, opt)
        if not ok:
            log.error("Unknown option %s" % opt)
            log.error("Full conf: %s" % conf)
        else:
            log.info("%s = %s" % (opt, val))

    def cmd_reset_config(self, option):
        if option == "":
            log.info("Resetting to defaults")
            self.callbacks["set_conf"](config.template_config)
        else:
            log.info("Resetting %s to defaults", option)
            conf = self.callbacks["get_conf"]()
            conf[option] = config.template_config[option]
            self.callbacks["set_conf"](conf)



# Favorite Plugin
# by Jack Miller
# v1.0

# This plugin allows certain items to be tagged as 'user:favorite' and then a
# custom style to be applied to those items.

from canto_next.plugins import check_program

check_program("canto-curses")

from canto_curses.taglist import TagListPlugin
from canto_curses.story import StoryPlugin
from canto_curses.command import register_commands

class StoryFavorite(StoryPlugin):
    def __init__(self, story):
        self.plugin_attrs = {}
        story.pre_format += "%?{'user:favorite' in ut}(*:)"

class TagListFavorite(TagListPlugin):
    def __init__(self, taglist):
        self.plugin_attrs = {}

        cmds = {
            "favorite" : (self.cmd_favorite, ["item-list"], "Favorite items."),
        }
        register_commands(taglist, cmds)

        self.taglist = taglist

        taglist.bind('*', 'favorite')

    def cmd_favorite(self, items):
        self.taglist.cmd_tag_item('%favorite', items)



# -*- coding: utf-8 -*-
#Canto-curses - ncurses RSS reader
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from canto_next.encoding import encoder, locale_enc
from .widecurse import waddch, wcwidth
from .html import html_entity_convert, char_ref_convert
from .config import config

import curses

import logging

log = logging.getLogger("WIDECURSE")

attr_count = { "B" : 0,
               "D" : 0,
               "R" : 0,
               "S" : 0,
               "U" : 0 }

attr_map = { "B" : curses.A_BOLD,
             "D" : curses.A_DIM,
             "R" : curses.A_REVERSE,
             "S" : curses.A_STANDOUT,
             "U" : curses.A_UNDERLINE }

# theme_print handles attribute codes and escaping:
#   %1 - %8 turns on color pairs 1 - 8
#   %0      turns on the previously enabled color

color_stack = []
color_stack_suspended = []

# Return length of next string of non-space characters
# or 0 if next character *is* a space.

def len_next_word(uni):
    if ' ' in uni:
        return theme_len(uni.split(' ', 1)[0])
    return theme_len(uni)

class FakePad():
    def __init__(self, width):
        self.x = 0
        self.y = 0
        self.width = width

    def attron(self, attr):
        pass

    def attroff(self, attr):
        pass

    def clrtoeol(self):
        pass

    def waddch(self, ch):
        cwidth = wcwidth(ch)
        if cwidth < 0 and not ch.is_space():
            return

        self.x += cwidth
        if self.x >= self.width:
            self.y += 1
            self.x -= self.width

    def getyx(self):
        return (self.y, self.x)

    def move(self, y, x):
        self.y = y
        self.x = x

class WrapPad():
    def __init__(self, pad):
        self.pad = pad

    def attron(self, attr):
        self.pad.attron(attr)

    def attroff(self, attr):
        self.pad.attroff(attr)

    def clrtoeol(self):
        self.pad.clrtoeol()

    def waddch(self, ch):
        waddch(self.pad, ch)

    def getyx(self):
        return self.pad.getyx()

    def move(self, x, y):
        return self.pad.move(x, y)

def theme_print_one(pad, uni, width):
    global color_stack
    global attr_count
    global attr_map

    max_width = width
    escaped = False
    code = False

    long_code = False
    lc = ""

    for i, c in enumerate(uni):
        ec = encoder(c)
        cwidth = wcwidth(ec)
        if cwidth < 0 and not ec.isspace():
            continue

        if escaped:
            # No room
            if cwidth > width:
                return "\\" + uni[i:]

            try:
                pad.waddch(ec)
            except:
                log.debug("Can't print escaped ec: %s in: %s", ec, uni)

            width -= cwidth
            escaped = False
        elif code:
            # Turn on color 1 - 8
            if c in "12345678":
                if len(color_stack):
                    pad.attroff(curses.color_pair(color_stack[-1]))
                color_stack.append(ord(c) - ord('0'))
                pad.attron(curses.color_pair(color_stack[-1]))
            # Return to previous color
            elif c == '0':
                if len(color_stack):
                    pad.attroff(curses.color_pair(color_stack[-1]))

                if len(color_stack) >= 2:
                    pad.attron(curses.color_pair(color_stack[-2]))
                    color_stack = color_stack[0:-1]
                else:
                    pad.attron(curses.color_pair(0))
                    color_stack = []

            # Turn attributes on / off
            elif c in "BbDdRrSsUu":
                if c.isupper():
                    attr_count[c] += 1
                else:
                    c = c.upper()
                    attr_count[c] -= 1

                if attr_count[c]:
                    pad.attron(attr_map[c])
                else:
                    pad.attroff(attr_map[c])

            # Suspend attributes
            elif c == "C":
                for attr in attr_map:
                    pad.attroff(attr_map[attr])
                for color in reversed(color_stack):
                    pad.attroff(curses.color_pair(color))
                pad.attron(curses.color_pair(0))
                color_stack_suspended = color_stack
                color_stack = []

            # Restore attributes
            elif c == "c":
                for attr in attr_map:
                    if attr_count[attr]:
                        pad.attron(attr_map[attr])
                color_stack = color_stack_suspended
                color_stack_suspended = []
                if color_stack:
                    pad.attron(curses.color_pair(color_stack[-1]))
                else:
                    pad.attron(curses.color_pair(0))
            elif c == "[":
                long_code = True
            code = False
        elif long_code:
            if c == "]":
                try:
                    long_color = int(lc)
                except:
                    log.error("Unknown long code: %s! Ignoring..." % lc)
                else:
                    if long_color < 1 or long_color > 256:
                        log.error("long color code must be >= 1 and <= 256")
                    else:
                        try:
                            pad.attron(curses.color_pair(long_color))
                            color_stack.append(long_color)
                        except:
                            log.error("Could not set pair. Perhaps need to set TERM='xterm-256color'?")
                long_code = False
                lc = ""
            else:
                lc += c
        elif c == "\\":
            escaped = True
        elif c == "%":
            code = True
        elif c == "\n":
            return uni[i + 1:]
        else:
            if c == " ":
                # Word too long
                wwidth = len_next_word(uni[i + 1:])

                # >= to account for current character
                if wwidth <= max_width and wwidth >= width:
                    return uni[i + 1:]

            # Character too long (should be handled above).
            if cwidth > width:
                return uni[i:]

            try:
                pad.waddch(ec)
            except Exception as e:
                log.debug("Can't print ec: %s in: %s", ec, repr(encoder(uni)))
                log.debug("Exception: %s", e)

            width -= cwidth

    return None

def theme_print(pad, uni, mwidth, pre = "", post = "", cursorbash=True, clear=True):
    prel = theme_len(pre)
    postl = theme_len(post)
    y = pad.getyx()[0]

    theme_print_one(pad, pre, prel)

    width = (mwidth - prel) - postl
    if width <= 0:
        raise Exception("theme_print: NO ROOM!")

    r = theme_print_one(pad, uni, width)

    if clear:
        pad.clrtoeol()

    if post:
        try:
            pad.move(y, (mwidth - postl))
        except:
            log.debug("move error: %d %d", y, mwidth - postl)
        theme_print_one(pad, post, postl)

    if cursorbash:
        try:
            pad.move(y + 1, 0)
        except:
            pass

    if r == uni:
        raise Exception("theme_print: didn't advance!")

    return r

# Returns the effective, printed length of a string, taking
# escapes and wide characters into account.

def theme_len(uni):
    escaped = False
    code = False
    length = 0

    for c in uni:
        ec = encoder(c)

        cwidth = wcwidth(ec)
        if cwidth < 0 and not ec.isspace():
            continue

        if escaped:
            length += cwidth
            escaped = False
        elif code:
            code = False
        elif c == "\\":
            escaped = True
        elif c == "%":
            code = True
        else:
            width = cwidth
            if width >= 0:
                length += width
    return length

# This is useful when a themed string needs to get truncated, so that color and
# attribute settings can be processed, despite the last part of the string not
# being displayed.

def theme_process(pad, uni):
    only_codes = ""
    escaped = False
    code = False

    for c in uni:
        if escaped:
            escaped = False
            continue
        elif code:
            only_codes += c
            code = False
        elif c == "\\":
            escaped = True
        elif c == "%":
            code = True
            only_codes += "%"

    # NOTE: len works because codes never use widechars.
    theme_print(pad, only_codes, len(only_codes), "", "", False)

# Strip more than two newlines from the front of the input, processing escapes
# as we discard characters.

def theme_lstrip(pad, uni):
    newlines = 0
    codes = ""
    escaped = False

    for i, c in enumerate(uni):
        # Discard
        if c in " \t\v":
            continue

        if c == "\n":
            newlines = 1
        elif c == "%":
            escaped = True
            codes += "%"
        elif escaped:
            escaped = False
            codes += c
        else:
            r = uni[i:]
            break

    # No content found.
    else:
        newlines = 0
        r = ""

    # Process dangling codes.
    if codes:
        theme_process(pad, codes)

    return (newlines * "\n") + r

def theme_reset():
    for key in attr_count:
        attr_count[key] = 0
    color_stack = []

utf_chars = { "ls" : "│",
              "rs" : "│",
              "ts" : "─",
              "bs" : "─",
              "tl" : "┌",
              "tr" : "┐",
              "bl" : "└",
              "br" : "┘" }

ascii_chars = { "ls" : "|",
                "rs" : "|",
                "ts" : "-",
                "bs" : "-",
                "tl" : "+",
                "tr" : "+",
                "bl" : "+",
                "br" : "+" }

def theme_border(code):
    if "UTF-8" in locale_enc:
        return utf_chars[code]
    return ascii_chars[code]

def prep_for_display(s):
    s = s.replace("\\", "\\\\")
    s = s.replace("%", "\\%")
    s = html_entity_convert(s)
    s = char_ref_convert(s)
    return s
